# -*- coding: utf-8 -*-
#
# badidatetime/badi_calendar.py
#
__docformat__ = "restructuredtext en"

import math

from badidatetime.base_calendar import BaseCalendar
from badidatetime.gregorian_calendar import GregorianCalendar
from badidatetime._coefficients import Coefficients

__all__ = ('BahaiCalendar',)


class BahaiCalendar(BaseCalendar, Coefficients):
    """
    Implementation of the Baha'i (Badi) Calendar.

    | WGS84--https://coordinates-converter.com/
    | https://whatismyelevation.com/location/35.63735,51.72569/Tehran--Iran-
    | https://en-us.topographic-map.com/map-g9q1h/Tehran/?center=35.69244%2C51.19492
    | https://www.google.com/maps/place/Tehran,+Tehran+Province,+Iran/@35.9098957,51.51371,9.49z/data=!4m6!3m5!1s0x3f8e02c69b919039:0x17c26479772c5928!8m2!3d35.6891975!4d51.3889736!16s%2Fm%2F025zk75?entry=ttu
    | https://gml.noaa.gov/grad/solcalc/ Sunset data
    """
    # Near Mehrabad International Airport
    #                  lattude    longitude  zone IANA name      elevation
    _BAHAI_LOCATION = (35.682376, 51.285817, 3.5, 'Asia/Tehran', 0)
    _GMT_LOCATION = (51.477928, -0.001545, 0, 0)
    _BADI_EPOCH = 2394643.5  # 2394645.5 using Meeus' algorithm
    _BADI_MONTH_NUM_DAYS = [
        (1, 19), (2, 19), (3, 19), (4, 19), (5, 19), (6, 19), (7, 19),
        (8, 19), (9, 19), (10, 19), (11, 19), (12, 19), (13, 19), (14, 19),
        (15, 19), (16, 19), (17, 19), (18, 19), (0, 0), (19, 19)
        ]
    KULL_I_SHAY_MIN = -5
    """
    Minimum kull_i_shay supported
    """
    KULL_I_SHAY_MAX = 4
    """
    Maximum kull_i_shay supported
    """
    MINYEAR = -1842
    """
    Minimum year supported
    """
    MAXYEAR = 1161
    """
    Maximum year supported
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        """
        kull_i_shay: 361-year (19^2) vahid (integer)
        vahid: (integer) 19-year vahid
        year: (integer) 1 - 19
        month: (integer) 1 - 19 plus 0 for Ayyām-i-Hā
        day: (integer) 1 - 19
        Baha'i long form date: [kull_i_shay, vahid, year, month, day]
        """
        self._bahai_date = None
        self._gc = GregorianCalendar()

    def utc_sunset(self, date: tuple, lat: float=None, lon: float=None,
                   zone: float=None) -> tuple:
        """
        Return the time of sunset in UTC time for the given Badi Day.

        :param tuple date: A Badi date.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The time zone.
        :returns: The hour, minute, and second of sunset based on the
                 provided coordinates.
        :rtype: tuple
        """
        jd = self.jd_from_badi_date(date[:3], lat, lon, zone)
        return self._hms_from_decimal_day(jd + 0.5)

    def naw_ruz_g_date(self, year: int, lat: float=None, lon: float=None,
                       zone: float=None, *, hms: bool=False) -> tuple:
        """
        Return the Badi date for Naw-Ruz from the given Badi year.

        :param int year: A Badi year.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The time zone.
        :param bool hms: If True the output returns the hours, minutes, and
                         seconds as seperate fields. If False the day has a
                         decimal value indicating the hours, minutes, and
                         seconds.
        :returns: A Gregorian date.
        :rtype: tuple
        """
        jd = self.jd_from_badi_date((year, 1, 1), lat, lon, zone)
        date = self._gc.gregorian_date_from_jd(jd, exact=True)
        return self._gc.ymdhms_from_date(date) if hms else date

    def first_day_of_ridvan_g_date(self, year: int, lat: float=None,
                                   lon: float=None, zone: float=None, *,
                                   hms: bool=False) -> tuple:
        """
        Find the first day of Riḍván in either with or without hours,
        minutes, and seconds.
        If the latitude, longitude, and time zone are not given Riḍván time
        of day is determined for the city of Nur in Iran.

        :param int year: A Badi year.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The time zone.
        :param bool hms: If True the output returns the hours, minutes, and
                         second as seperate fields. If False the day has a
                         decimal value indicating the hours, minutes, and
                         seconds.
        :returns: A Gregorian date.
        :rtype: tuple
        """
        jd = self.jd_from_badi_date((year, 2, 13), lat, lon, zone)
        date = self._gc.gregorian_date_from_jd(jd, exact=True)
        return self._gc.ymdhms_from_date(date) if hms else date

    def jd_from_badi_date(self, b_date: tuple, lat: float=None,
                          lon: float=None, zone: float=None,
                          _chk_on: bool=True) -> float:
        """
        Convert a Badi short form date to Julian period day.

        :param tuple b_date: A short form Badi date.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The time zone.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The Julian Period day.
        :rtype: float
        """
        year, month, day = self.date_from_kvymdhms(
            self.long_date_from_short_date(b_date, trim=True, _chk_on=_chk_on),
            short=True, _chk_on=_chk_on)

        if month == 0:  # Ayyam-i-Ha
            days = 18 * 19
        elif month < 19:  # month 1 - 18
            days = (month - 1) * 19
        else:  # month == 19:
            days = 18 * 19 + 4 + self._is_leap_year(year, _chk_on=_chk_on)

        td = self._days_in_years(year-1)
        jd = td + math.floor(self._BADI_EPOCH+1) + days + day

        if any([True if l is None else False for l in (lat, lon, zone)]):
            lat, lon, zone = self._BAHAI_LOCATION[:3]

        # The diff value converts the more exact jd to the Meeus algorithm
        # for determining the sunset jd. The fractional on the day is not
        # affected.
        diff = self._meeus_from_exact(jd)
        ss_a = self._sun_setting(jd + diff, lat, lon, zone) % 1
        return round(jd + ss_a + self._get_jd_coeff(year),
                     self._ROUNDING_PLACES)

    def _get_jd_coeff(self, year: int) -> int:
        """
        Generate the coefficients for correcting Badí' vernal equinox dates.

        .. note::

           | General ranges are determined with:
           | ./contrib/misc/badi_jd_tests.py -p -S start_year -E end_year

           Where -S is the 1st year and -E is the nth year + 1 that needs to
           be process. Use the following command to test the results of each
           segment.
           ./contrib/misc/badi_jd_tests.py -qXS start_year -E end_year

           Full range is -1842 to 1161.

        :param int year: The year to find a coefficient for.
        :returns: The coefficient.
        :rtype: int
        """
        def process_segment(y, a=0, onoff0=(), b=0, onoff1=()):
            func = lambda y, onoff: 0 < y < 100 and y % 4 in onoff
            coeff = 0

            if a and func(y, onoff0):    # Whatever is passed in onoff0.
                coeff = a
            elif b and func(y, onoff1):  # Whatever is passed in onoff1.
                coeff = b

            return coeff

        def process_segments(year, pn, a=0, onoff0=(), b=0, onoff1=()):
            coeff = 0

            for start, end in pn:
                if year in range(start, end):
                    # Start to end (range -S start -E end)
                    coeff0 = process_segment(end - year, a=a, onoff0=onoff0)
                    coeff1 = process_segment(end - year, b=b, onoff1=onoff1)
                    coeff = coeff0 if coeff0 != 0 else coeff1

            return coeff

        coeff = 0

        if not coeff:
            coeff = process_segments(year, self._P1, -1, (0, 1, 2, 3))

        if not coeff:
            coeff = process_segments(year, self._P1100, -1, (0, 3))

        if not coeff:
            coeff = process_segments(year, self._P1110, -1, (0, 2, 3))

        if not coeff:
            coeff = process_segments(year, self._P2, -2, (0, 1, 2, 3))

        if not coeff:
            coeff = process_segments(year, self._P2111, -2, (0,), -1,
                                     (1, 2, 3))

        if not coeff:
            coeff = process_segments(year, self._P2211, -2, (0, 3), -1,
                                     (1, 2))

        if not coeff:
            coeff = process_segments(year, self._P2221, -2, (0, 2, 3), -1,
                                     (1,))

        return coeff

    def badi_date_from_jd(self, jd: float, lat: float=None, lon: float=None,
                          zone: float=None, *, us: bool=False,
                          short: bool=False, fraction: bool=False,
                          trim: bool=False, rtd: bool=False,
                          _chk_on: bool=True) -> tuple:
        """
        Convert a Julian Period day to a Badi date.

        :param float jd: Julian Period day.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The standard time zone.
        :param bool us: If True the seconds are split to seconds amd
                        microseconds else if False the seconds has a partial
                        day as a decimal.
        :param bool short: If True then parse for a short date else if False
                           (default) parse for a long date.
        :param bool fraction: This will return a short date with a possible
                              fraction on the day.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :param bool rtd: Round to day.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The Badi date from a Julian Period day.
        :rtype: tuple
        """
        def get_leap_year_info(year, _chk_on):
            leap = self._is_leap_year(year, _chk_on=_chk_on)
            yds = 365 + leap
            ld = 4 + leap
            return leap, yds, ld

        def check_and_fix_day(cjd, y, lat=None, lon=None, zone=None,
                              _chk_on=True):
            fjdy = self.jd_from_badi_date((y, 1, 1), lat, lon, zone, _chk_on)
            return y-1 if (math.floor(fjdy) - math.floor(cjd)) > 0 else y

        md = jd - (self._BADI_EPOCH - 1)
        # This is only needed for the last two days of Badi year 1161
        # and the day before the epoch.
        y = 1 if md < 424046 and md != 0 else 0
        year = math.floor(md / self._MEAN_TROPICAL_YEAR) + y
        leap, yds, ld = get_leap_year_info(year, _chk_on)

        if (y := check_and_fix_day(jd, year, lat, lon, zone, _chk_on)):
            year = y
            leap, yds, ld = get_leap_year_info(year, _chk_on)

        fjdy = self.jd_from_badi_date((year, 1, 1), lat, lon, zone,
                                      _chk_on=_chk_on)
        days = math.floor(jd) - math.floor(fjdy) + 1

        if days <= 342:  # Month 1 - 18
            m_days = days % 19
            day = 19 if m_days == 0 else m_days
        elif (342 + ld) < days <= yds:  # Month 19
            day = days - (342 + ld)
        else:  # Ayyam-i-Ha
            day = days % 342

        month_days = self._BADI_MONTH_NUM_DAYS
        month_days[18] = (0, ld)  # Fix Ayyám-i-Há days

        for month, ds in month_days:
            if days <= ds: break
            days -= ds

        if any([True if l is None else False for l in (lat, lon, zone)]):
            lat, lon, zone = self._BAHAI_LOCATION[:3]

        date = self._adjust_date(jd, (year, month, day), lat, lon, zone,
                                 fraction=fraction, rtd=rtd)

        if fraction:
            b_date = date
        else:
            l_date = self.long_date_from_short_date(date, trim=True,
                                                    _chk_on=_chk_on)
            b_date = self.kvymdhms_from_b_date(l_date, us=us, short=short,
                                               trim=trim, _chk_on=_chk_on)

        return b_date

    def short_date_from_long_date(self, b_date: tuple, *, trim: bool=False,
                                  _chk_on: bool=True) -> tuple:
        """
        Convert a long date (kvymdhms) to a short date (ymdhms). In either
        case microseconds could also be provided.

        :param tuple b_date: A long form date with or without microseconds.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The short form Badi date.
        :rtype: tuple
        """
        kull_i_shay, vahid, year, month, day = b_date[:5]
        hh, mm, ss, us = self._get_hms(b_date)
        y = (kull_i_shay - 1) * 361 + (vahid - 1) * 19 + year
        hmsms = self._trim_hms((hh, mm, ss, us)) if trim else (hh, mm, ss, us)
        date = (y, month, day) + hmsms
        _chk_on and self._check_valid_badi_date(date, short_in=True)
        return date

    def long_date_from_short_date(self, date: tuple, *, trim: bool=False,
                                  _chk_on: bool=True) -> tuple:
        """
        Convert a date to a short date (ymdhms) to a long date (kvymdhms).

        :param tuple b_date: A short form date with or without microseconds.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :param bool _chk_on: If True (default) all date check are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The long form Badi date.
        :rtype: tuple
        """
        y, month, day = date[:3]
        hh, mm, ss, us = self._get_hms(date, short_in=True)
        k = y / 361
        kull_i_shay = 0 if y == 0 else math.ceil(k)
        k0 = self._truncate_decimal(k % 1, self._ROUNDING_PLACES)
        # k0 = round(k % 1, self._ROUNDING_PLACES)
        v = k0 / 19 * 361

        if v == 0:  # If there is no fraction in v
            vahid = 19
            year = 19
        else:
            vahid = math.ceil(v)
            year = math.ceil(v % 1 * 19)

        hmsms = self._trim_hms((hh, mm, ss, us)) if trim else (hh, mm, ss, us)
        b_date = (kull_i_shay, vahid, year, month, day) + hmsms
        _chk_on and self._check_valid_badi_date(b_date)
        return b_date

    def date_from_kvymdhms(self, b_date: tuple, *, short: bool=False,
                           _chk_on: bool=True) -> tuple:
        """
        Convert (Kull-i-Shay, Váḥid, year, month, day, hour, minute, second,
        us) into a (Kull-i-Shay, Váḥid, year, month, day.fraction) or
        (year, month, day.fraction) date.

        :param tuple b_date: The Badi date in long form.
        :param bool short: If True then parse for a short date else if False
                           (default) parse for a long date.
        :param bools _chk_on: If True (default) all date checks are enforced
                              else if False they are turned off. This is only
                              used internally. Do not use unless you know what
                              you are doing.
        :returns: The long or short form Badi date with hours, minutes,
                  seconds, and microseconds if set.
        :rtype: tuple
        """
        _chk_on and self._check_valid_badi_date(b_date)
        kull_i_shay, vahid, year, month, day = b_date[:5]
        hour, minute, second, us = self._get_hms(b_date)
        day += round(self._HR(hour) + self._MN(minute) + self._SEC(second) +
                     self._US(us), self._ROUNDING_PLACES)
        date = (kull_i_shay, vahid, year, month, day)
        return (self.short_date_from_long_date(date, trim=True, _chk_on=_chk_on)
                if short else date)

    def kvymdhms_from_b_date(self, b_date: tuple, *, us: bool=False,
                             short: bool=False, trim: bool=False,
                             _chk_on: bool=True) -> tuple:
        """
        Convert (Kull-i-Shay, Váḥid, year, month, day.fraction) into
        (Kull-i-Shay, Váḥid, year, month, day, hour, minute, second) or if
        short is True (year, month, day, hour, minute, second). If us is
        True the seconds are split to second and microsecond.

        :param tuple b_date: The Badi date in long form.
        :param bool us: If True the seconds are split to seconds amd
                        microseconds else if False the seconds has a partial
                        day as a decimal.
        :param bool short: If True then parse for a short date else if False
                           (default) parse for a long date.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The long or short form Badi date with hours, minutes,
                  seconds, and microseconds if set.
        :rtype: tuple
        """
        dlen = len(b_date)

        # We need to trim any zero hh, mm, ss, us so partial days below
        # work correctly.
        if dlen > 5:
            hms = self._trim_hms(b_date[5:dlen])
            b_date = b_date[:5] + hms
            dlen = len(b_date)

        _chk_on and self._check_valid_badi_date(b_date)
        kull_i_shay, vahid, year, month, day = b_date[:5]

        if dlen == 5:
            hd = self._PARTIAL_DAY_TO_HOURS(day)
            hour = math.floor(hd)
            md = self._PARTIAL_HOUR_TO_MINUTE(hd)
            minute = math.floor(md)
            second = round(self._PARTIAL_MINUTE_TO_SECOND(md),
                           self._ROUNDING_PLACES)
        else:
            hour = b_date[5] if dlen > 5 else 0
            minute = b_date[6] if dlen > 6 else 0
            second = b_date[7] if dlen > 7 else 0

        date = (kull_i_shay, vahid, year, month, math.floor(day))

        if us:
            hmsms = (hour, minute, *self._sec_microsec_from_seconds(second))
        else:
            hmsms = (hour, minute, second)

        date += self._trim_hms(hmsms) if trim else hmsms
        return (self.short_date_from_long_date(date, trim=trim, _chk_on=_chk_on)
                if short else date)

    def badi_date_from_gregorian_date(self, g_date: tuple, lat: float=None,
                                      lon: float=None, zone: float=None, *,
                                      short: bool=False, trim: bool=False,
                                      rtd: bool=False, _exact: bool=True
                                      ) -> tuple:
        """
        Get the Badi date from the Gregorian date.

        :param tuple g_date: A Gregorian date.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The standard time zone.
        :param bool short: If True then parse for a short date else if False
                           (default) parse for a long date.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :param bool rtd: Round to day.
        :param bool _exact: Use the more exact Julian Period algorithm.
                            Default is True. This should generally be set to
                            True, a False value will give inaccurate results
                            and is used for testing only.
        :returns: A Badi date long or short form.
        :rtype: tuple
        """
        jd = self._gc.jd_from_gregorian_date(g_date, exact=_exact)
        return self.badi_date_from_jd(jd, lat=lat, lon=lon, zone=zone,
                                      short=short, trim=trim, rtd=rtd)

    def gregorian_date_from_badi_date(self, b_date: tuple, lat: float=None,
                                      lon: float=None, zone: float=None, *,
                                      _exact: bool=True, _chk_on: bool=True
                                      ) -> tuple:
        """
        Get the Gregorian date from the Badi date.

        :param tuple b_date: A Badi date short form.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The standard time zone.
        :param bool _exact: Use the more exact Julian Period algorithm.
                            Default is True. This should generally be set to
                            True, a False value, in this method will give
                            inaccurate results and is used for testing only.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The Gregorian date.
        :rtype: tuple
        """
        jd = self.jd_from_badi_date(b_date, lat, lon, zone, _chk_on=_chk_on)
        return self._gc.ymdhms_from_date(self._gc.gregorian_date_from_jd(
            jd, exact=_exact), us=True)

    def posix_timestamp(self, t: float, lat: float=None, lon: float=None,
                        zone: float=None, *, us: bool=False, short: bool=False,
                        trim: bool=False) -> tuple:
        """
        Get the Badi date from a POSIX timestamp.

        :param float t: Timestamp
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The time zone.
        :param bool us: If True the seconds are split to seconds amd
                        microseconds else if False the seconds has a partial
                        day as a decimal.
        :param bool short: If True then parse for a short date else if False
                           (default) parse for a long date.
        :param bool trim: Trim the us, ss, mm, and hh in that order.
        :returns: A Badi date long or short form.
        :rtype: tuple
        """
        jd = t / 86400 + self._POSIX_EPOCH
        return self.badi_date_from_jd(jd, lat, lon, zone, us=us, short=short,
                                      trim=trim)

    def midday(self, date: tuple, *, hms: bool=False, _chk_on: bool=True
               ) -> tuple:
        """
        Find the midday time in hours, minutes, and seconds with fraction.

        :param tuple date: Badi date short or long.
        :param bool hms: If True return the hours, minutes, and seconds else
                         if False return the decimal value.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: Midday in hours, minutes, and seconds.
        :rtype: tuple
        """
        if len(date) == 5:
            b_date = self.short_date_from_long_date(date, trim=True,
                                                    _chk_on=_chk_on)
        else:
            b_date = date

        jd = self.jd_from_badi_date(b_date)
        diff0 = self._meeus_from_exact(jd)
        ss0 = self._sun_setting(jd + diff0, *self._GMT_LOCATION[:3])
        diff1 = self._meeus_from_exact(jd + 1)
        ss1 = self._sun_setting(jd + 1 + diff1, *self._GMT_LOCATION[:3])
        mid = (ss1 - ss0) / 2
        return self._hms_from_decimal_day(mid) if hms else mid

    def _trim_hms(self, hms: tuple) -> tuple:
        """
        Trim the hours, minutes, seconds or microseconds off the date if
        zero unless a lower value was not zero.

        .. list-table:: Examples
           :widths: 18 16 66
           :header-rows: 1

           * - Examples
             - Results
             - Description
           * - (12, 30, 6, 0)
             - (12, 30, 6)
             - The zero microseconds would be trimmed.
           * - (12,  0, 6, 0)
             - (12, 0, 6)
             - The zero microseconds would be trimmed but the zero minutes
               would be left untouched.

        :param tuple hms: An hour, minute, and second object.
        :returns: An object with the lower order parts stripped off if
                  they have a zero value.
        :rtype: tuple
        """
        items = []
        has = False

        for v in reversed(hms):
            if v == 0 and not has:
                continue
            else:
                items.append(v)
                has = True

        return tuple(reversed(items))

    def _check_valid_badi_date(self, b_date: tuple, short_in: bool=False
                               ) -> None:
        """
        Check that the Kull-i-Shay, Váḥids, year, month, day, hour, minute,
        second, and microsecond values are valid.

        :param tuple b_date: A long form Badi date.
        :param bool short_in: If True then parse for a short date else if
                              False parse for a long date. This is for
                              incoming dates not outgoing dates as in most
                              other uses of 'short'.
        :returns: Nothing
        :rtype: None
        :raises AssertionError: When a date Váḥid, year, month, day, hour,
                                minute, second, or microsecond are out of
                                range.
        """
        cycle = 20

        if not short_in:  # Long Badi date
            kull_i_shay, vahid, year, month, day = b_date[:5]
            hour, minute, second, us = self._get_hms(b_date)
            assert (self.KULL_I_SHAY_MIN <= kull_i_shay
                    <= self.KULL_I_SHAY_MAX), (
                f"Invalid kull-i-shay {kull_i_shay}, it must be in the range "
                f"of [{self.KULL_I_SHAY_MIN}, {self.KULL_I_SHAY_MAX}].")
            assert 1 <= vahid < cycle, (
                f"Invalid Váḥids '{vahid}' in a Kull-i-Shay’, it must be in "
                "the range of [1, 19].")
            assert 1 <= year < cycle, (
                f"Invalid year '{year}' in a Váḥid, it must be in the "
                "range of [1, 19].")
            ly = (kull_i_shay - 1) * 361 + (vahid - 1) * 19 + year
        else:  # Short Badi date
            year, month, day = b_date[:3]
            hour, minute, second, us = self._get_hms(b_date, short_in=True)
            assert self.MINYEAR <= year <= self.MAXYEAR, (
                f"Invalid year '{year}' it must be in the range of ["
                f"{self.MINYEAR}, {self.MAXYEAR}].")
            ly = year

        assert 0 <= month < cycle, (
            f"Invalid month '{month}', it must be in the range of [0, 19].")
        # This is Ayyām-i-Hā and could be 4 or 5 days depending on leap year.
        cycle = (5 + self._is_leap_year(ly)) if month == 0 else cycle
        assert 1 <= day < (cycle), (
            f"Invalid day '{day}' for month '{month}', it must be in the "
            f"range of [1, {cycle-1}].")
        self._check_valid_badi_time(hour, minute, second, us)

        # Check if there are any fractionals that invalidate other values.
        if any((hour, minute, second)):
            assert not day % 1, (
                "If there is a part day then there can be no hours, minutes, "
                "or seconds.")

        if any((minute, second)):
            assert not hour % 1, ("If there is a part hour then there can "
                                  "be no minutes or seconds.")

        if second:
            assert not minute % 1, (
                "If there is a part minute then there can be no seconds.")

    def _check_valid_badi_time(self, hour: float, minute: float, second: float,
                               us: int, maxsec: int=60) -> None:
        """
        Check that the hour, minute, second, and microsecond values are valid.

        :param float hour: Hours
        :param float minute: Minutes
        :param float second: Seconds
        :param float us: Microseconds
        :returns: Nothing
        :rtype: None
        :raises AssertionError: When an hour, minute, second, or microsecond
                                are out of range.
        """
        assert 0 <= hour < 25, (
            f"Invalid hour '{hour}', it must be in the range of [0, 24].")
        assert 0 <= minute < 60, (
            f"Invalid minute '{minute}', it must be in the range of [0, 59].")
        assert 0 <= second < maxsec, (
            f"Invalid second '{second}', it must be in the range of "
            f"[0, {maxsec}].")
        assert 0 <= us < 1000000, (
            f"Invalid microseconds '{us}', it must be in the range of "
            "[0, 999999].")

    def _is_leap_year(self, year: int, _chk_on: bool=True) -> bool:
        """
        Return a Boolean True if a Badí' leap year, False if not.

        :param int year: This value must be a Badí' short form year.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: A Boolean indicating if a leap year or not.
        :rtype: bool
        """
        return self._days_in_year(year, _chk_on=_chk_on) == 366

    def _days_in_year(self, year: int, _chk_on: bool=True) -> int:
        """
        Determine the number of days in the provided Badi year.

        :param int year: The Badi year to process.
        :param bool _chk_on: If True (default) all date checks are enforced
                             else if False they are turned off. This is only
                             used internally. Do not use unless you know what
                             you are doing.
        :returns: The number of days.
        :rtype: int
        """
        jd_n0 = self.jd_from_badi_date((year, 1, 1), _chk_on=_chk_on)
        # For year 1162 we need to turn off the date check so we can get
        # the leap year for 1161.
        on = False if (year + 1) == 1162 else True
        jd_n1 = self.jd_from_badi_date((year + 1, 1, 1), _chk_on=on)
        return int(math.floor(jd_n1) - math.floor(jd_n0))

    def _get_hms(self, date: tuple, *, short_in: bool=False) -> tuple:
        """
        Parse the hours, minutes, seconds, and microseconds, if they exist
        for either the short or long form Badi date.

        :param tuple date: A long or short form Badi date.
        :param bool short_in: If True then parse for a short date else if False
                              parse for a long date. This is for incoming dates
                              not outgoing dates as in most other uses of
                              'short'.
        :returns: The relevant hours, minutes, and seconds.
        :rtype: tuple
        """
        t_len = len(date)
        s = 3 if short_in else 5
        hour = date[s] if t_len > s and date[s] is not None else 0
        minute = date[s+1] if t_len > s+1 and date[s+1] is not None else 0
        second = date[s+2] if t_len > s+2 and date[s+2] is not None else 0
        us = date[s+3] if t_len > s+3 and date[s+3] is not None else 0
        return hour, minute, second, us

    def _adjust_date(self, jd: float, ymd: tuple, lat: float, lon: float,
                     zone: float, *, fraction: bool=False, us: bool=False,
                     rtd: bool=False) -> tuple:
        """
        The adjusted year, month, and day based on if the JD falls before
        or after sunset.

        .. warning::

           This method will give disasterous results if the jd and ymd
           arguments are off by more than a day. As of now this method does
           not work with a JD after noon on the following day.

        :param float jd: Exact Julian Period day.
        :param tuple ymd: The year month, and day.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The standard time zone.
        :param bool fraction: If True output the day with a fractional day as
                              a decimal. If False (default) output hour,
                              minute, and second.
        :param bool us: If True convert a fractional second to microseconds. If
                        False (default) output second with a fraction.
        :param bool rtd: Round to day.
        :returns: Returns the year, month, day, and depending on other
                  arguments, the hour, minute, and second, and microsecond.
        :rtype: tuple
        """
        assert self._xor_boolean((fraction, us, rtd)), (
            "Cannot set more than one of fraction, us, or rtd to True.")
        year, month, day = ymd
        jd0 = math.floor(jd)  # So we always get the sunset on the current day.
        mjd0 = jd0 + self._meeus_from_exact(jd0)  # Current day
        # Current day sunset
        ss0 = self._sun_setting(mjd0, lat, lon, zone)
        ss_frac = round(ss0 % 1, self._ROUNDING_PLACES)
        jd_frac = round(jd % 1, self._ROUNDING_PLACES)

        if jd_frac < ss_frac:  # Previous day before sunset.
            jd1 = jd0 - 1
            mjd1 = jd1 + self._meeus_from_exact(jd1)
            ss1 = self._sun_setting(mjd1, lat, lon, zone)
            ss1_frac = round(ss1 % 1, self._ROUNDING_PLACES)
            frac = abs(0.5 - ss1_frac + jd_frac)
            day -= 1

            if day == 0:
                if month == 1:  # Stage 1
                    year -= 1
                    month = 19
                    day = 19
                    # print('Stage 1', jd, ymd, jd_frac, ss_frac, ss1 % 1)
                elif month in range(2, 19):  # Stage 2
                    month -= 1
                    day = 19
                    # print('Stage 2', jd, ymd, jd_frac, ss_frac, ss1 % 1)
                elif month == 19:  # Stage 3
                    month = 0
                    day = 4 + self._is_leap_year(year)
                    # print('Stage 3', jd, ymd, jd_frac, ss_frac, ss1 % 1)
                else:  # month 0 -> Ayyám-i-Há Stage 4
                    month = 18
                    day = 19
                    # print('Stage 4', jd, ymd, jd_frac, ss_frac, ss1 % 1)
        elif jd_frac <= 0.5:  # Same Badi day before or equal to UTC midnight.
            frac = jd_frac - ss_frac
            # print('Stage 5', jd, ymd, jd_frac, ss_frac)
        elif 0.5 < jd_frac:  # Same Badi day after UTC midnight -- Stage 6
            frac = 0.5 - ss_frac + (jd_frac - 0.5)
            # print('Stage 6', jd, ymd, ss_frac, jd_frac)
        else:  # pragma: no cover -- Stage 7
            assert False, (f"Should never happen--jd: {jd}, ymd: {ymd}, "
                           f"jd_frac: {jd_frac}, ss_frac: {ss_frac}")

        if fraction:
            day = round(day + frac, self._ROUNDING_PLACES)
            hms = ()
        elif rtd:
            day = round(day + frac)
            hms = ()
        else:
            hh, mm, ss = self._hms_from_decimal_day(frac)

            if us:
                microsecond = self._PARTIAL_SECOND_TO_MICROSECOND(ss)
                ss = math.floor(ss)
                msec = (microsecond,)
            else:
                ss = round(ss, self._ROUNDING_PLACES)
                msec = ()

            hms = (hh, mm, ss) + msec

        return (year, month, day) + hms

    def _day_Length(self, jd: float, lat: float, lon: float, zone: float
                    ) -> tuple:
        """
        The hour, minute, and seconds of the day's offset either less than
        or more than 24 hours.

        :param float jd: The astronomically exact Julian Period day.
        :param float lat: The latitude.
        :param float lon: The longitude.
        :param float zone: The standard time zone.
        :returns: The hour, minute, and second.
        :rtype: tuple
        """
        jd0 = math.floor(jd)
        jd1 = jd0 + 1
        # The next day
        mjd1 = jd1 + self._meeus_from_exact(jd1)
        ss1 = self._sun_setting(mjd1, lat, lon, zone)
        # The first day
        mjd0 = jd0 + self._meeus_from_exact(jd0)
        ss0 = self._sun_setting(mjd0, lat, lon, zone)
        # Subtract the first day from the next day gived the total
        # hours, minutes, and seconds between them.
        value = list(self._hms_from_decimal_day(ss1 - ss0))
        value[0] = 24 if value[0] == 0 else value[0]
        return tuple(value)



# -*- coding: utf-8 -*-
#
# badidatetime/test/test_strptime.py
#
__docformat__ = "restructuredtext en"

import os
import re
import unittest
import importlib
import locale
from unittest.mock import patch

from .._timedateutils import _td_utils
from .._strptime import (_getlang, LocaleTime, _calc_julian_from_U_or_W,
                         DotDict, StrpTime, TimeRE, _strptime_time,
                         _strptime_datetime)

datetime = importlib.import_module('badidatetime.datetime')


class TestStrptime_Functions(unittest.TestCase):

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        self.getlocale_patcher = patch(
            'badidatetime._strptime.locale.getlocale')
        self.mock_getlocale = self.getlocale_patcher.start()
        self.mock_getlocale.return_value = ('en_US', 'UTF-8')

    def tearDown(self):
        self.getlocale_patcher.stop()

    #@unittest.skip("Temporarily skipped")
    def test__getlang(self):
        """
        Test that the _getlang function returns the current language.
        """
        result = _getlang()
        expected = ('en_US', 'UTF-8')
        msg = (f"Expected {expected}, found {result}.")
        self. assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test__calc_julian_from_U_or_W(self):
        """
        Test that the _calc_julian_from_U_or_W function returns the day of the
        year.
        The week_of_year is from 0 to 50/51
        The day_of_week is from 0 to 6.
        The 1st day is always week 0.
        """
        data = (
            # year, week_of_year, day_of_week, expected_result
            # 22   6    Istijlāl Kamál      False 365  152
            (-1, 21, 5, 152),  # (-1, 8, 19)
            # 01   4    Fiḍāl    Bahá       True  366  001
            (1, 0, 3, 1),      # (1, 1, 1)
            # 52   5    `Idāl    Bahá       True  366  001
            (182, 0, 4, 1),    # (182, 1, 1)
            # 52   6    Istijlāl Bahá       True  366  002
            (182, 0, 5, 2),    # (182, 1, 2)
            # 52   7    Istiqlāl Bahá       True  366  003
            (182, 0, 6, 3),    # (182, 1, 3)
            # 01   1    Jalál    Bahá       True  366  004
            (182, 0, 0, 4),    # (182, 1, 4)
            )
        msg = "Expected {}, with year {}, WoY {}, and DoW {}, found {}."

        for year, woy, dow, expected_result in data:
            result = _calc_julian_from_U_or_W(year, woy, dow)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, year, woy, dow, result))

    #@unittest.skip("Temporarily skipped")
    def test__strptime_time(self):
        """
        Test that the _strptime_time function returns the correct time struct
        depending on the format data.
        """
        data = (
            ('Jal Bah 05 12:30:30 1', '',
             'structures.ShortFormStruct(tm_year=1, tm_mon=1, tm_mday=5, '
             'tm_hour=12, tm_min=30, tm_sec=30, tm_wday=0, tm_yday=5, '
             'tm_isdst=0)'),
            ('0182-01-16', '%Y-%m-%d',
             'structures.ShortFormStruct(tm_year=182, tm_mon=1, tm_mday=16, '
             'tm_hour=0, tm_min=0, tm_sec=0, tm_wday=5, tm_yday=16, '
             'tm_isdst=0)'),
            )
        msg = "Expected {}, found {}"

        for data_string, fmt, expected_result in data:
            if fmt == '':
                result = _strptime_time(data_string)
            else:
                result = _strptime_time(data_string, fmt)

            self.assertEqual(expected_result, str(result), msg.format(
                expected_result, result))

    #@unittest.skip("Temporarily skipped")
    def test__strptime_datetime(self):
        """
        Test that the _strptime_datetime function returns an instance of
        the supplyed datetime class object.
        """
        def safe_data_parse():
            is_github = os.getenv('TEST_RUNNING') == 'true'

            if not is_github:
                extra_tests = (
                    ('182-01-16T12:30:30 EST', '%Y-%m-%dT%H:%M:%S %Z',
                     '0182-01-16T12:30:30'),
                    ('182-01-16T12:30:30-0500 EST', '%Y-%m-%dT%H:%M:%S%z %Z',
                     '0182-01-16T12:30:30-05:00'),
                    )
            else:
                extra_tests = ()
                print(f"GitHub running tests. (TEST_RUNNING == {is_github})")

            return extra_tests

        data = (
            ('Jal Bah 05 12:30:30 1', '', '0001-01-05T12:30:30'),
            ('0182-01-16', '%Y-%m-%d', '0182-01-16T00:00:00'),
            ('0 5 12:30:30', '%m %d %H:%M:%S', '0001-00-05T12:30:30'),
            ('12:30:30', '%H:%M:%S', '0002-01-01T12:30:30'),
            ('182-01-16T12:30:30-05:00', '%Y-%m-%dT%H:%M:%S%z',
             '0182-01-16T12:30:30-05:00'),
            )
        msg = "Expected {}, found {}"
        data += safe_data_parse()

        for cnt, (data_string, fmt, expected_result) in enumerate(data):
            try:
                if fmt == '':
                    result = _strptime_datetime(datetime.datetime, data_string)
                else:
                    result = _strptime_datetime(datetime.datetime, data_string,
                                                fmt)

                self.assertEqual(expected_result, str(result), msg.format(
                    expected_result, result))
            except ValueError as e:
                raise ValueError(f"With cnt '{cnt}' an error was raised, {e}")


class TestStrptime_LocaleTime(unittest.TestCase):
    """
    This test is a bit different than most test. Since all the methods in
    the class are called in the constructor we need to test the results of
    each of the methods.
    """

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        os.environ['TZ'] = 'EST+05EDT,M4.1.0,M10.5.0'
        self._lt = LocaleTime()
        self._lt.LC_date_time = 'kam jam  17 22:44:30 0199'
        self._lt.LC_date = '%m/%d/%Y'
        self._lt.LC_time = '%I:%M:%S'

    #@unittest.skip("Temporarily skipped")
    def test_a_weekday(self):
        """
        Test that the abbreviated weekdays exist ard are in the correct order.
        """
        expected = [_td_utils.DAYNAMES_ABV[i].lower() for i in range(7)]
        result = self._lt.a_weekday
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_f_weekday(self):
        """
        Test that the full weekdays exist ard are in the correct order.
        """
        expected = [_td_utils.DAYNAMES[i].lower() for i in range(7)]
        result = self._lt.f_weekday
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_a_month(self):
        """
        Test that the abbreviated months exist ard are in the correct order.
        """
        expected = [_td_utils.MONTHNAMES_ABV[i].lower() for i in range(20)]
        result = self._lt.a_month
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_f_month(self):
        """
        Test that the full months exist ard are in the correct order.
        """
        expected = [_td_utils.MONTHNAMES[i].lower() for i in range(20)]
        result = self._lt.f_month
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_am_pm(self):
        """
        Test that the am_pm designators are properly set.
        """
        expected = ['am', 'pm']
        result = self._lt.am_pm
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_LC_date_time(self):
        """
        Test that the LC_date_time variable is set properly.
        """
        expected = 'kam jam  17 22:44:30 0199'
        result = self._lt.LC_date_time
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_LC_date(self):
        """
        Test that the LC_date variable is set properly.
        """
        expected = '%m/%d/%Y'
        result = self._lt.LC_date
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_LC_time(self):
        """
        Test that the LC_time variable is set properly.
        """
        expected = '%I:%M:%S'
        result = self._lt.LC_time
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_timezone(self):
        """
        Test that the timezone variable is set properly.
        """
        expected = (frozenset({'gmt', 'est', 'utc'}), frozenset({'edt'}))
        result = self._lt.timezone
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)


class TestStrptime_TimeRE(unittest.TestCase):

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        self._tre = TimeRE()
        self._tre.locale_time.LC_date_time = '%a %b  %d %H:%M:%S %Y'
        self._tre.locale_time.LC_date = '%m/%d/%Y'
        self._tre.locale_time.LC_time = '%I:%M:%S'

    #@unittest.skip("Temporarily skipped")
    def test_regex_set_from_constructor(self):
        """
        Test that the constructor set up 27 items in the custom dict.
        """
        expected = 27
        result = len(self._tre)
        msg = "Expected {}, found {}"
        self.assertEqual(expected, result, msg.format(expected, result))

    #@unittest.skip("Temporarily skipped")
    def test___seqToRE(self):
        """
        Test that the __seqToRE method returns an empty string when the
        `to_convert` argument has empty values.
        """
        to_convert = ('',)
        result = self._tre._TimeRE__seqToRE(to_convert, '')
        expected = ''
        msg = "Expected {}, found {}"
        self.assertEqual(expected, result, msg.format(expected, result))

    #@unittest.skip("Temporarily skipped")
    def test_pattern(self):
        """
        Test that the pattern method returns a regex pattern.
        """
        data = (
            (self._tre.locale_time.LC_date_time,
             r"(?P<a>jal|jam|kam|fiḍ|idā|isj|isq)\s+(?P<b>ayy|bah|jal|jam|"
             r"aẓa|núr|raḥ|kal|kam|asm|izz|mas|ilm|qud|qaw|mas|sha|sul|mul|"
             r"alá)\s+(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])\s+(?P<H>2[0-3]|[0-1]"
             r"\d|\d):(?P<M>[0-5]\d|\d):(?P<S>6[0-1]|[0-5]\d|\d)\s+(?P<Y>-?"
             r"\d{1,4})"),
            (self._tre.locale_time.LC_date,
             r"(?P<m>1[0-9]|0[0-9]|[0-9])/(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])"
             r"/(?P<Y>-?\d{1,4})"),
            (self._tre.locale_time.LC_time,
             r"(?P<I>1[0-2]|0[1-9]|[1-9]):(?P<M>[0-5]\d|\d):(?P<S>6[0-1]|"
             r"[0-5]\d|\d)"),
            )
        msg = "Expected {}, found {}"

        for fmt, expected_result in data:
            result = self._tre.pattern(fmt)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, result))


class TestStrptime_DotDict(unittest.TestCase):

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        aritrary_dict = {1: 1, 'a': 'a'}
        self._dd = DotDict(aritrary_dict)

    def test_get(self):
        """
        Test that the get method returns the correct object from the
        dictionary.
        """
        err_msg0 = "No such attribute: {}"
        data = (
            (1, False, 1),
            ('a', False, 'a'),
            ('b', True, err_msg0.format('b')),
            )
        msg = "Expected {}, found {}"

        for key, validity, expected_result in data:
            if validity:
                try:
                    result = getattr(self._dd, key)
                except AttributeError as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    # Raise an error when an AssertionError is not raised.
                    raise AssertionError(f"With key '{key}' an error was not "
                                         f"raised, returned {result}.")
            else:
                result = self._dd.get(key)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, result))

    def test_set(self):
        """
        Test that the set method creates a new object in the dictionary.
        """
        key = 'c'
        expected = 'c'
        self._dd[key] = 'c'
        result = getattr(self._dd, key)
        msg = f"Expected {expected}, found {result}"
        self.assertEqual(expected, result, msg)


class TestStrptime_StrpTime(unittest.TestCase):

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        self.locale_patcher = patch(
            'badidatetime._timedateutils.locale.nl_langinfo')
        self.mock_nl_langinfo = self.locale_patcher.start()

        def side_effect(item):
            if item == locale.LC_TIME:
                return 'en_US.UTF-8'
            elif item == locale.AM_STR:
                return 'AM'
            elif item == locale.PM_STR:
                return 'PM'
            elif item == locale.D_FMT:
                return '%m/%d/%Y'
            elif item == locale.T_FMT:
                return '%H:%M:%S'
            else:
                return 'Default Value'

        self.mock_nl_langinfo.side_effect = side_effect

    def tearDown(self):
        self.locale_patcher.stop()

    #@unittest.skip("Temporarily skipped")
    def test_constructor_exception(self):
        """
        """
        with self.assertRaises(TypeError) as cm:
            StrpTime(100)

        message = str(cm.exception)
        err_msg = "strptime() argument 0 must be str, not <class 'int'>."
        self.assertEqual(err_msg, message)

    #@unittest.skip("Temporarily skipped")
    def test__find_regex(self):
        """
        Test that the _strptime function returns a complex tuple in the form
        of ((year, month, day, None, None,
             hour, minute, second,
             weekday, julian, tz, tzname, gmroff), fraction, gmtoff_fraction)
        """
        err_msg0 = "'{}' is a bad directive in format '{}'"
        err_msg1 = "stray %% in format '{}'"
        err_msg2 = "Time data '{}' does not match format '{}'."
        err_msg3 = "Unconverted data remains: {}"
        data = (
            # The default format does not work in my code or the standard
            # Python code.
            ('Jal Bah 01 06:12:30 182', '%a %b %d %H:%M:%S %Y', False,
             "<re.Match object; span=(0, 23), match='Jal Bah "
             "01 06:12:30 182'>"),
            ('06:12:30', '%H:%M:%S', False,
             "<re.Match object; span=(0, 8), match='06:12:30'>"),
            ('1', '%U', False, "<re.Match object; span=(0, 1), match='1'>"),
            ('52', '%V', False, "<re.Match object; span=(0, 2), match='52'>"),
            ('10', '% ', True, err_msg0.format(
                '%', '<built-in function format>')),
            ('12', '%', True, err_msg1.format('%')),
            ('stuff', '%y', True, err_msg2.format('stuff', '%y')),
            ('123', '%u', True, err_msg3.format('23')),
            )
        msg = "Expected {}, found {}"

        for data_str, fmt, validity, expected_result in data:
            st = StrpTime(data_str, fmt)
            st._clear_cache()

            if validity:
                try:
                    st._find_regex()
                except ValueError as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    # Raise an error when an AssertionError is not raised.
                    raise AssertionError(f"With data_str {data_str} an error "
                                         "was not raised.")
            else:
                result = st._find_regex()
                self.assertEqual(expected_result, str(result[0]), msg.format(
                    expected_result, result))

    #@unittest.skip("Temporarily skipped")
    def test__parse_found_dict(self):
        """
        Test that the _parse_found_dict method correctly parses the arguments
        using the correct regex.
        """
        err_msg0 = "Inconsistent use of : in {}"
        data = (
            (r"(?P<y>\d\d)", 'year', '42', False, 42),
            (r"(?P<Y>-?\d\d\d\d)", 'year', '0042', False, 42),
            (r"(?P<G>-?\d\d\d\d)", 'iso_year', '0181', False, 181),
            (r"(?P<G>-?\d\d\d\d)", 'iso_year', '-1842', False, -1842),
            (r"(?P<m>1[0-9]|0[1-9]|[1-9])", 'month', '19', False, 19),
            (r"(?P<m>1[0-9]|0[1-9]|[1-9])", 'month', '09', False, 9),
            (r"(?P<m>1[0-9]|0[1-9]|[1-9])", 'month', '9', False, 9),
            (r"(?P<m>1[0-9]|0[0-9]|[0-9])", 'month', '00', False, 0),
            (r"(?P<B>ayyám\-i\-há|mashíyyat|'aẓamat|kalimát|masá’il|raḥmat|"
             r"'izzat|qudrat|sharaf|sulṭán|jalál|jamál|kamál|asmá'|'alá'|bahá|"
             r"'ilm|qawl|mulk|núr)", 'month', 'Bahá', False, 1),
            (r"(?P<B>ayyám-i-há|mashíyyat|'aẓamat|kalimát|masá’il|raḥmat|"
             r"'izzat|qudrat|sharaf|sulṭán|jalál|jamál|kamál|asmá'|'alá'|bahá|"
             r"'ilm|qawl|mulk|núr)", 'month', 'Ayyám-i-Há', False, 0),
            (r"(?P<b>ayy|bah|jal|jam|aẓa|núr|raḥ|kal|kam|asm|izz|mas|ilm|qud|"
             r"qaw|mas|sha|sul|mul|alá)", 'month', 'Alá', False, 19),
            (r"(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])", 'day', '09', False, 9),
            (r"(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])", 'day', '9', False, 9),
            (r"(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])", 'day', ' 9', False, 9),
            (r"(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])", 'day', '19', False, 19),
            (r"(?P<H>2[0-3]|[0-1]\d|\d)", 'hour', '23', False, 23),
            (r"(?P<H>2[0-3]|[0-1]\d|\d)", 'hour', '02', False, 2),
            (r"(?P<H>2[0-3]|[0-1]\d|\d)", 'hour', '2', False, 2),
            (r"(?P<I>1[0-2]|0[1-9]|[1-9])", 'hour', '12', False, 0),  # am
            (r"(?P<I>1[0-2]|0[1-9]|[1-9])", 'hour', '02', False, 2),
            (r"(?P<I>1[0-2]|0[1-9]|[1-9])", 'hour', '2', False, 2),
            (r"(?P<M>[0-5]\d|\d)", 'minute', '00', False, 0),
            (r"(?P<M>[0-5]\d|\d)", 'minute', '9', False, 9),
            (r"(?P<S>6[0-1]|[0-5]\d|\d)", 'second', '60', False, 60),
            (r"(?P<S>6[0-1]|[0-5]\d|\d)", 'second', '59', False, 59),
            (r"(?P<S>6[0-1]|[0-5]\d|\d)", 'second', '9', False, 9),
            (r"(?P<f>[0-9]{1,6})", 'fraction', '09', False, 90000),
            (r"(?P<A>istijlāl|istiqlāl|jalál|jamál|kamál|fiḍāl|`idāl)",
             'weekday', 'Jalál', False, 0),
            (r"(?P<A>istijlāl|istiqlāl|jalál|jamál|kamál|fiḍāl|`idāl)",
             'weekday', '`Idāl', False, 4),
            (r"(?P<a>jal|jam|kam|fiḍ|idā|isj|isq)", 'weekday', 'Jal', False, 0),
            (r"(?P<a>jal|jam|kam|fiḍ|idā|isj|isq)", 'weekday', 'Idā', False, 4),
            (r"(?P<w>[0-6])", 'weekday', '6', False, 6),
            (r"(?P<u>[1-7])", 'weekday', '7', False, 7),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '365', False, 365),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '350', False, 350),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '250', False, 250),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '090', False, 90),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '009', False, 9),
            (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]\d|"
             r"0[1-9]|[1-9])", 'julian', '9', False, 9),
            (r"(?P<U>5[0-2]|[0-4]\d|\d)", 'week_of_year', '52', False, 52),
            (r"(?P<U>5[0-2]|[0-4]\d|\d)", 'week_of_year', '40', False, 40),
            (r"(?P<U>5[0-2]|[0-4]\d|\d)", 'week_of_year', '01', False, 1),
            (r"(?P<W>5[0-2]|[0-4]\d|\d)", 'week_of_year', '52', False, 52),
            (r"(?P<W>5[0-2]|[0-4]\d|\d)", 'week_of_year', '40', False, 40),
            (r"(?P<W>5[0-2]|[0-4]\d|\d)", 'week_of_year', '01', False, 1),
            (r"(?P<V>5[0-3]|0[1-9]|[1-4]\d|\d)", 'iso_week', '52', False, 52),
            (r"(?P<Z>gmt|est|utc|edt)", 'tz', 'GMT', False, 0),
            (r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",
             'gmtoff', '-0500', False, -18000),
            (r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",
             'gmtoff', '-05:00:00', False, -18000),
            (r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",
             'gmtoff', 'Z', False, 0),  # Zulu or GMT
            (r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",
             'gmtoff', '-03:3015', True, err_msg0.format('-03:3015')),
            )
        msg = "Expected {}, with cnt {} and data_str {}, found {}"

        for cnt, (regex, variable, data_str, validity,
                  expected_result) in enumerate(data):
            st = StrpTime(data_str)
            locale_time = LocaleTime()
            cmp_regex = re.compile(regex, re.IGNORECASE)
            found = cmp_regex.match(data_str)

            if validity:
                try:
                    st._parse_found_dict(found, locale_time)
                except ValueError as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    # Raise an error when an AssertionError is not raised.
                    raise AssertionError(f"With count test {cnt} an error "
                                         "was not raised.")
            else:
                dot_dict = st._parse_found_dict(found, locale_time)
                result = getattr(dot_dict, variable)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, cnt, data_str, result))

    #@unittest.skip("Temporarily skipped")
    def test_multi_char_formats(self):
        """
        Test that the _parse_found_dict method correctly parses multi
        character formats using the correct regex.
        """
        data = (
            (r"(?P<I>1[0-2]|0[1-9]|[1-9])\s*(?P<p>am|pm)", '2 pm',
             {'hour': 14}),
            (r"(?P<y>\d\d) %", '01 %', {'year': 1, }),
            (r"(?P<a>jal|jam|kam|fiḍ|idā|isj|isq)\s+(?P<b>ayy|bah|jal|jam|"
             r"aẓa|núr|raḥ|kal|kam|asm|izz|mas|ilm|qud|qaw|mas|sha|sul|mul|"
             r"alá)\s+(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])\s+(?P<H>2[0-3]|"
             r"[0-1]\d|\d):(?P<M>[0-5]\d|\d):(?P<S>6[0-1]|[0-5]\d|\d)"
             r"\s+(?P<Y>-?\d{1,4})", 'Idā Bah 01 12:30:30 182',
             {'year': 182, 'month': 1, 'day': 1, 'hour': 12, 'minute': 30,
              'second': 30}),
            (r"(?P<m>1[0-9]|0[0-9]|[0-9])/(?P<d>1[0-9]|0[1-9]|[1-9]|"
             r" [1-9])/(?P<Y>-?\d{1,4})", '19/19/0181',
             {'year': 181, 'month': 19, 'day': 19}),
            (r"(?P<I>1[0-2]|0[1-9]|[1-9]):(?P<M>[0-5]\d|\d):(?P<S>6[0-1]|"
             r"[0-5]\d|\d)", '11:30:30',
             {'hour': 11, 'minute': 30, 'second': 30}),
            )
        msg = "Expected {}, with cnt {} and data_str {}, found {}"

        for cnt, (regex, data_str, expected_result) in enumerate(data):
            st = StrpTime(data_str)
            locale_time = LocaleTime()
            cmp_regex = re.compile(regex, re.IGNORECASE)
            found = cmp_regex.match(data_str)

            try:
                dot_dict = st._parse_found_dict(found, locale_time)
            except AttributeError as e:
                raise AttributeError(f"Count {cnt}, {e}")

            for key, expected in expected_result.items():
                result = getattr(dot_dict, key)
                self.assertEqual(expected, result, msg.format(
                    expected_result, cnt, data_str, result))

    #@unittest.skip("Temporarily skipped")
    def test__check_iso_week(self):
        """
        Test that the _check_iso_week method raises the appropreate exceptions.
        """
        err_msg0 = ("Day of the year directive '%j' is not compatible with "
                    "ISO year directive '%G'. Use '%Y' instead.")
        err_msg1 = ("ISO year directive '%G' must be used with the ISO week "
                    "directive '%V' and a weekday directive ('%A', '%a', "
                    "'%w', or '%u').")
        err_msg2 = ("ISO week directive '%V' must be used with the ISO year "
                    "directive '%G' and a weekday directive ('%A', '%a', "
                    "'%w', or '%u').")
        err_msg3 = ("ISO week directive '%V' is incompatible with the year "
                    "directive '%Y'. Use the ISO year '%G' instead.")
        data = (
            ('iso_year', None, 'julian', None, 'iso_week', None,
             'weekday', '2', 'year', '-1842', False, ''),
            ('iso_year', '-1842', 'julian', None, 'iso_week', '1',
             'weekday', '2', 'year', '-1842', False, ''),
            ('iso_year', '-1842', 'julian', '1', 'iso_week', '1',
             'weekday', '2', 'year', '-1842', True, err_msg0),
            ('iso_year', '-1842', 'julian', None, 'iso_week', None,
             'weekday', '2', 'year', '-1842', True, err_msg1),
            ('iso_year', '-1842', 'julian', None, 'iso_week', '1',
             'weekday', None, 'year', '-1842', True, err_msg1),
            ('iso_year', None, 'julian', None, 'iso_week', '1',
             'weekday', '1', 'year', None, True, err_msg2),
            ('iso_year', None, 'julian', None, 'iso_week', '1',
             'weekday', None, 'year', '182', True, err_msg2),
            ('iso_year', None, 'julian', None, 'iso_week', '1',
             'weekday', '1', 'year', '182', True, err_msg3),
            )

        for cnt, (var0, value0, var1, value1, var2, value2, var3, value3,
                  var4, value4, validity, expected_result) in enumerate(data):
            st = StrpTime("")
            dot_dict = DotDict({var0: value0, var1: value1,
                                var2: value2, var3: value3, var4: value4})

            if validity:
                try:
                    st._check_iso_week(dot_dict)
                except ValueError as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    # Raise an error when an AssertionError is not raised.
                    raise AssertionError(f"With count test {cnt} an error "
                                         "was not raised.")
            else:  # Valid tests (Nothing to assert)
                st._check_iso_week(dot_dict)

    #@unittest.skip("Temporarily skipped")
    def test__miscellaneous(self):
        """
        Test that the _miscellaneous method updates a few values.

        Both `weekday` and `week_of_year` start their count with 0 not 1.
        """
        data = (
            # Setting the julian day (day of the year) if it is None and
            # weekday is not None.
            ({'julian': None, 'weekday': 4, 'week_of_year': 0, 'year': 182},
             {'julian': 1}),
            # Updating the julian day (day of the year).
            ({'julian': None, 'weekday': 6, 'week_of_year': 0, 'year': 182,
              'iso_year': None, 'iso_week': None}, {'julian': 3}),
            # Setting the year, month, day, and julian
            ({'julian': None, 'weekday': 4, 'week_of_year': None, 'year': None,
              'month': None, 'iso_year': 182, 'iso_week': 1},
             {'year': 182, 'month': 1, 'day': 8, 'julian': 8, 'weekday': 4}),
            # Setting the year, month, and day.
            ({'julian': 1, 'weekday': None, 'year': 182},
             {'year': 182, 'month': 1, 'day': 1, 'weekday': 4}),
            # Test for leap year when Ayyám-i-Há 5 is given.
            ({'julian': None, 'weekday': None, 'year': None, 'month': 0,
              'day': 5}, {'year': 1}),
            ({'julian': None, 'weekday': 0, 'week_of_year': -1, 'year': 182},
             {'julian': 355, 'year': 181}),  # 0181-19-09
            )
        msg = "Expected {}, with variable '{}', found {}"

        for cnt, (items, expected_result) in enumerate(data):
            dd = DotDict(items)
            st = StrpTime("")

            try:
                st._miscellaneous(dd)
            except AttributeError as e:
                raise AttributeError(f"Test count {cnt}, {e}.")

            for var, expected in expected_result.items():
                result = getattr(dd, var)
                self.assertEqual(expected, result, msg.format(
                    expected, var, result))



# -*- coding: utf-8 -*-
#
# badidatetime/test/test_julian_period.py
#
__docformat__ = "restructuredtext en"

import unittest

from ..julian_period import JulianPeriod
from ..gregorian_calendar import GregorianCalendar


class TestJulianPeriod(unittest.TestCase):

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        self._jp = JulianPeriod()
        self._gc = GregorianCalendar()

    #@unittest.skip("Temporarily skipped")
    def test__julian_centuries(self):
        """
        Test that the _julian_centuries method returns the Julian
        century in dynamical time from a Julian moment.
        """
        data = (
            (2394646.5, -1.5577960301163587),  # Badi epoch
            (1721425.5, -19.98958247775496),
            (self._jp._J2000, 0.0),
            )
        msg = "Expected {} for jd {}, found {}"

        for jd, expected_result in data:
            result = self._jp._julian_centuries(jd)
            self.assertEqual(expected_result, result,
                             msg.format(expected_result, jd, result))

    #@unittest.skip("Temporarily skipped")
    def test__julian_millennia(self):
        """
        Test that the julian_millennis method returns the Julian
        millennia in dynamical time from a Julian moment.
        """
        data = (
            (2394646.5, -0.15577960301163587),
            (1721425.5, -1.9989582477754961),
            (self._jp._J2000, 0.0),
            )
        msg = "Expected {} for jd {}, found {}"

        for jd, expected_result in data:
            result = self._jp._julian_millennia(jd)
            self.assertEqual(expected_result, result,
                             msg.format(expected_result, jd, result))



# -*- coding: utf-8 -*-
#
# badidatetime/_strptime.py
#
__docformat__ = "restructuredtext en"

"""
Strptime-related classes and functions.

CLASSES:
    LocaleTime -- Discovers and stores locale-specific time information
    TimeRE -- Creates regexes for pattern matching a string of text containing
                time information
    StrpTime -- Calculates the time struct represented by the passed-in string

FUNCTIONS:
    _getlang -- Figure out what language is being used for the locale
"""

import time
import locale
from re import compile as re_compile
from re import IGNORECASE
from re import escape as re_escape
from _thread import allocate_lock as _thread_allocate_lock

from .datetime import (date as datetime_date,
                       timedelta as datetime_timedelta,
                       timezone as datetime_timezone)
from ._timedateutils import _td_utils
from .badi_calendar import BahaiCalendar


__all__ = []


def _getlang():
    # Figure out what the current language is set to.
    return locale.getlocale(locale.LC_TIME)


class LocaleTime:
    """
    Stores and handles locale-specific information related to time.

    ATTRIBUTES:
        f_weekday -- full weekday names (7-item list)
        a_weekday -- abbreviated weekday names (7-item list)
        f_month -- full month names (13-item list; dummy value in [0], which
                    is added by code)
        a_month -- abbreviated month names (13-item list, dummy value in
                    [0], which is added by code)
        am_pm -- AM/PM representation (2-item list)
        LC_date_time -- format string for date/time representation (string)
        LC_date -- format string for date representation (string)
        LC_time -- format string for time representation (string)
        timezone -- daylight- and non-daylight-savings timezone representation
                    (2-item list of sets)
        lang -- Language used by instance (2-item tuple)
    """

    def __init__(self):
        """
        Set all attributes.

        Order of methods called matters for dependency reasons.

        The locale language is set at the offset and then checked again before
        exiting.  This is to make sure that the attributes were not set with a
        mix of information from more than one locale.  This would most likely
        happen when using threads where one thread calls a locale-dependent
        function while another thread changes the locale while the function in
        the other thread is still running.  Proper coding would call for
        locks to prevent changing the locale while locale-dependent code is
        running.  The check here is done in case someone does not think about
        doing this.

        Only other possible issue is if someone changed the timezone and did
        not call tz.tzset .  That is an issue for the programmer, though,
        since changing the timezone is worthless without that call.
        """
        self.lang = _getlang()
        self.__calc_weekday()
        self.__calc_month()
        self.__calc_am_pm()
        self.__calc_timezone()
        self.__calc_date_time()

        if _getlang() != self.lang:  # pragma: no cover
            raise ValueError("locale changed during initialization")

        if (time.tzname != self.tzname or
            time.daylight != self.daylight):  # pragma: no cover
            raise ValueError("timezone changed during initialization")

    def __calc_weekday(self):
        """
        Set self.a_weekday and self.f_weekday using the _timedateutils module.
        """
        self.a_weekday = [_td_utils.DAYNAMES_ABV[i].lower() for i in range(7)]
        self.f_weekday = [_td_utils.DAYNAMES[i].lower() for i in range(7)]

    def __calc_month(self):
        """
        Set self.f_month and self.a_month using the _timedateutils module.
        """
        month_names = [v for k, v in sorted(_td_utils.MONTHNAMES_ABV.items(),
                                            key=lambda k: k)]
        self.a_month = [month_names[i].lower() for i in range(20)]
        month_names = [v for k, v in sorted(_td_utils.MONTHNAMES.items(),
                                            key=lambda k: k)]
        self.f_month = [month_names[i].lower() for i in range(20)]

    def __calc_am_pm(self):
        """
        Set self.am_pm by using time.strftime().

        The magic date (1, 1 , 1, hour, 0, 0, 0, 0, 0) is not really that
        magical, I just needed a time shortly after sunset and a time shortly
        before the next sunset.
        """
        am_pm = []

        for hour in (1, 22):
            # The short form Badí' date is used.
            time_tuple = _td_utils._build_struct_time(
                (199, 1, 1, hour, 44, 55), 0, short_in=True)
            am_pm.append(_td_utils.strftime("%p", time_tuple).lower())

        self.am_pm = am_pm

    def __calc_date_time(self):
        # Set self.date_time, self.date, & self.time by using
        # time.strftime().

        # Use (199, 3, 17, 22, 44, 55, 4, 1, 0) for magic date because the
        # amount of overloaded numbers is minimized. The order in which
        # searches for values within the format string is very important;
        # it eliminates possible ambiguity for what something represents.
        time_tuple = _td_utils._build_struct_time((199, 3, 17, 22, 4, 30), 0,
                                                  short_in=True)
        date_time = [None, None, None]
        date_time[0] = _td_utils.strftime("%c", time_tuple).lower()
        date_time[1] = _td_utils.strftime("%x", time_tuple).lower()
        date_time[2] = _td_utils.strftime("%X", time_tuple).lower()
        # date_time = ['kam jam  17 22:44:30 0199', '03/17/0199', '10:44:30']
        replacement_pairs = [('%', '%%'), (self.f_weekday[2], '%A'),
                             (self.f_month[3], '%B'), (self.a_weekday[2], '%a'),
                             (self.a_month[3], '%b'), (self.am_pm[1], '%p'),
                             ('0199', '%Y'), ('99', '%y'), ('22', '%H'),
                             ('04', '%M'), ('4', '%-M'), ('30', '%S'),
                             ('55', '%j'), ('17', '%d'), ('03', '%m'),
                             ('3', '%-m'), ('2', '%w'), ('10', '%I')]
        replacement_pairs.extend([(tz, "%Z") for tz_values in self.timezone
                                  for tz in tz_values])

        for offset, directive in ((0, '%c'), (1, '%x'), (2, '%X')):
            current_format = date_time[offset]

            for old, new in replacement_pairs:
                # Must deal with possible lack of locale info
                # manifesting itself as the empty string (e.g., Swedish's
                # lack of AM/PM info) or a platform returning a tuple of empty
                # strings (e.g., MacOS 9 having timezone as ('','')).
                if old:
                    current_format = current_format.replace(old, new)

            time_tuple = _td_utils._build_struct_time((199, 1, 1, 1, 1, 1), 0,
                                                      short_in=True)
            date_time[offset] = current_format.replace('11', '%U')

        self.LC_date_time = date_time[0]
        self.LC_date = date_time[1]
        self.LC_time = date_time[2]

    def __calc_timezone(self):
        # Set self.timezone by using time.tzname.
        # Do not worry about possibility of time.tzname[0] == time.tzname[1]
        # and time.daylight; handle that in strptime.
        try:
            time.tzset()
        except AttributeError:  # pragma: no cover
            pass

        self.tzname = time.tzname
        self.daylight = time.daylight
        no_saving = frozenset({"utc", "gmt", self.tzname[0].lower()})

        if self.daylight:
            has_saving = frozenset({self.tzname[1].lower()})
        else:  # pragma: no cover
            has_saving = frozenset()

        self.timezone = (no_saving, has_saving)


class TimeRE(dict):
    """
    Handle conversion from format directives to regexes.
    """

    def __init__(self, locale_time=None):
        """
        Create keys/values.

        Order of execution is important for dependency reasons.
        """
        if locale_time:  # pragma: no cover
            self.locale_time = locale_time
        else:
            self.locale_time = LocaleTime()

        base = super()
        base.__init__({
            # The " [1-9]" part of the regex is to make %c from ANSI C work
            'd': r"(?P<d>1[0-9]|0[1-9]|[1-9]| [1-9])",
            'f': r"(?P<f>[0-9]{1,6})",
            'H': r"(?P<H>2[0-3]|[0-1]\d|\d)",
            'I': r"(?P<I>1[0-2]|0[1-9]|[1-9])",
            'G': r"(?P<G>-?\d\d\d\d)",
            'j': (r"(?P<j>36[0-6]|3[0-5]\d|[1-2]\d\d|0[1-9]\d|00[1-9]|[1-9]"
                  r"\d|0[1-9]|[1-9])"),
            'm': r"(?P<m>1[0-9]|0[0-9]|[0-9])",
            'M': r"(?P<M>[0-5]\d|\d)",
            'S': r"(?P<S>6[0-1]|[0-5]\d|\d)",
            'U': r"(?P<U>5[0-2]|[0-4]\d|\d)",
            'w': r"(?P<w>[0-6])",
            'u': r"(?P<u>[1-7])",
            'V': r"(?P<V>5[0-3]|0[1-9]|[1-4]\d|\d)",
            # W is set below by using 'U'
            'y': r"(?P<y>\d\d)",
            'Y': r"(?P<Y>-?\d{1,4})",  # We alow negative years.
            'z': r"(?P<z>[+-]\d\d:?[0-5]\d(:?[0-5]\d(\.\d{1,6})?)?|(?-i:Z))",
            'A': self.__seqToRE(self.locale_time.f_weekday, 'A'),
            'a': self.__seqToRE(self.locale_time.a_weekday, 'a'),
            'B': self.__seqToRE(self.locale_time.f_month, 'B'),
            'b': self.__seqToRE(self.locale_time.a_month, 'b'),
            'p': self.__seqToRE(self.locale_time.am_pm, 'p'),
            'Z': self.__seqToRE((tz for tz_names in self.locale_time.timezone
                                 for tz in tz_names), 'Z'),
            '%': '%'})
        base.__setitem__('W', base.__getitem__('U').replace('U', 'W'))
        base.__setitem__('c', self.pattern(self.locale_time.LC_date_time))
        base.__setitem__('x', self.pattern(self.locale_time.LC_date))
        base.__setitem__('X', self.pattern(self.locale_time.LC_time))

    def __seqToRE(self, to_convert, directive):
        """
        Convert a list to a regex string for matching a directive.

        We want possible matching values to be from longest to shortest. This
        prevents the possibility of a match occurring for a value that also
        a substring of a larger value that should have matched (e.g., 'abc'
        matching when 'abcdef' should have been the match).
        """
        to_convert = sorted(to_convert, key=len, reverse=True)

        for value in to_convert:
            if value != '':
                break
        else:
            return ''

        regex = '|'.join(re_escape(stuff) for stuff in to_convert)
        regex = f'(?P<{directive}>{regex}'
        return f'{regex})'

    def pattern(self, format):
        """
        Return regex pattern for the format string.

        Need to make sure that any characters that might be interpreted as
        regex syntax are escaped.
        """
        processed_format = ''
        # The sub() call escapes all characters that might be misconstrued
        # as regex syntax.  Cannot use re.escape since we have to deal with
        # format directives (%m, etc.).
        regex_chars = re_compile(r"([\\.^$*+?\(\){}\[\]|])")
        format = regex_chars.sub(r"\\\1", format)
        whitespace_replacement = re_compile(r'\s+')
        format = whitespace_replacement.sub(r'\\s+', format)

        while '%' in format:
            directive_index = format.index('%') + 1
            processed_format = (f"{processed_format}"
                                f"{format[:directive_index - 1]}"
                                f"{self[format[directive_index]]}")
            format = format[directive_index + 1:]

        return f"{processed_format}{format}"

    def compile(self, format):
        """
        Return a compiled re object for the format string.
        """
        return re_compile(self.pattern(format), IGNORECASE)


_cache_lock = _thread_allocate_lock()
# DO NOT modify _TimeRE_cache or _regex_cache without acquiring the cache lock
# first!
_TimeRE_cache = TimeRE()
_CACHE_MAX_SIZE = 5  # Max number of regexes stored in _regex_cache
_regex_cache = {}


def _calc_julian_from_U_or_W(year, week_of_year, day_of_week):
    """
    Calculate the Julian day based on the year, week of the year, and day of
    the week, with week_start_day representing whether the week of the year
    assumes the week starts on Bahá (Saturday) 0. This means that both %U and
    %W would give the same results.

    The `week_of_year` is NOT based on the ISO standard. If it's the 1st day
    of the year then it is always week 0.
    """
    first_weekday = datetime_date(year, 1, 1).weekday()
    # Need to watch out for a week 0 (when the first day of the year is not
    # the same as that specified by %U or %W).
    week_0_length = (7 - first_weekday) % 7

    if week_of_year == 0:
        if day_of_week < first_weekday:
            ret = 7 - first_weekday + day_of_week + 1
        else:
            ret = 1 + day_of_week - first_weekday
    else:
        days_to_week = week_0_length + (7 * (week_of_year - 1))
        ret = 1 + days_to_week + day_of_week

    return ret


class DotDict(dict):
    """
    Allows dot notation access to dictionary attributes.
    """
    def __init__(self, dictionary):
        super().__init__(dictionary)

        for key, value in dictionary.items():
            self[key] = DotDict(value) if isinstance(value, dict) else value

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(f"No such attribute: {key}")

    def __setattr__(self, key, value):
        self[key] = value


class StrpTime:
    """
    Return a 2-tuple consisting of a time struct and an int containing
    the number of microseconds based on the input string and the
    format string.
    """

    def __init__(self, data_string, format="%a %b %d %H:%M:%S %Y"):
        for index, arg in enumerate([data_string, format]):
            if not isinstance(arg, str):
                msg = "strptime() argument {} must be str, not {}."
                raise TypeError(msg.format(index, type(arg)))

        self.data_string = data_string
        self.format = format

    def start(self):
        found, locale_time = self._find_regex()
        dot_dict = self._parse_found_dict(found, locale_time)
        self._check_iso_week(dot_dict)
        self._miscellaneous(dot_dict)
        # Always returns a short form Badi date--for now.
        return ((dot_dict.year, dot_dict.month, dot_dict.day,
                 dot_dict.hour, dot_dict.minute, dot_dict.second,
                 dot_dict.weekday, dot_dict.julian, dot_dict.tz,
                 dot_dict.tzname, dot_dict.gmtoff),
                dot_dict.fraction, dot_dict.gmtoff_fraction)

    def _clear_cache(self):
        global _regex_cache
        _regex_cache.clear()

    def _find_regex(self):
        global _TimeRE_cache, _regex_cache

        with _cache_lock:
            locale_time = _TimeRE_cache.locale_time

            if (_getlang() != locale_time.lang or
                time.tzname != locale_time.tzname or
                time.daylight != locale_time.daylight):  # pragma: no cover
                _TimeRE_cache = TimeRE()
                self._clear_cache()
                locale_time = _TimeRE_cache.locale_time

            if len(_regex_cache) > _CACHE_MAX_SIZE:  # pragma: no cover
                self._clear_cache()

            format_regex = _regex_cache.get(self.format)

            if not format_regex:
                try:
                    format_regex = _TimeRE_cache.compile(self.format)
                # KeyError raised when a bad format is found; can be specified
                # as \\, in which case it was a stray % but with a space after
                # it.
                except KeyError as err:
                    bad_directive = err.args[0]

                    if bad_directive == "\\":
                        bad_directive = "%"

                    del err
                    raise ValueError(f"'{bad_directive}' is a bad directive "
                                     f"in format '{format}'") from None
                # IndexError only occurs when the format string is "%"
                except IndexError:
                    raise ValueError(f"stray %% in format '{self.format}'"
                                     ) from None

                _regex_cache[format] = format_regex

        found = format_regex.match(self.data_string)

        if not found:
            raise ValueError(f"Time data {self.data_string!r} does not match "
                             f"format {self.format!r}.")

        if len(self.data_string) != found.end():
            raise ValueError("Unconverted data remains: "
                             f"{self.data_string[found.end():]}")

        return found, locale_time

    def _parse_found_dict(self, found, locale_time):
        values = {'iso_year': None, 'year': None, 'month': 1, 'day': 1,
                  'hour': 0, 'minute': 0, 'second': 0, 'fraction': 0, 'tz': -1,
                  'gmtoff': None, 'gmtoff_fraction': 0, 'iso_week': None,
                  'week_of_year': None, 'weekday': None, 'julian': None}
        # Weekday and julian defaulted to None so as to signal need to
        # calculate values.
        dot_dict = DotDict(values)
        found_dict = found.groupdict()

        for group_key in found_dict.keys():
            # Directives not explicitly handled below:
            #   c, x, X
            #      handled by making out of other directives
            #   U, W
            #      worthless without day of the week
            if group_key == 'y':
                # With the year without a century an accurate year cannot
                # be derived, since the min and max years supported is from
                # -1842 to 1161.
                dot_dict.year = int(found_dict['y'])
            elif group_key == 'Y':
                dot_dict.year = int(found_dict['Y'])
            elif group_key == 'G':
                dot_dict.iso_year = int(found_dict['G'])
            elif group_key == 'm':
                dot_dict.month = int(found_dict['m'])
            elif group_key == 'B':
                dot_dict.month = locale_time.f_month.index(
                    found_dict['B'].lower())
            elif group_key == 'b':
                dot_dict.month = locale_time.a_month.index(
                    found_dict['b'].lower())
            elif group_key == 'd':
                dot_dict.day = int(found_dict['d'])
            elif group_key == 'H':
                dot_dict.hour = int(found_dict['H'])
            elif group_key == 'I':
                dot_dict.hour = int(found_dict['I'])
                ampm = found_dict.get('p', '').lower()

                # If there was no AM/PM indicator, we'll treat this like AM.
                if ampm in ('', locale_time.am_pm[0]):
                    # We're in AM so the hour is correct unless we're
                    # looking at 12 midnight.
                    # 12 midnight == 12 AM == hour 0
                    if dot_dict.hour == 12:
                        dot_dict.hour = 0
                elif ampm == locale_time.am_pm[1]:
                    # We're in PM so we need to add 12 to the hour unless
                    # we're looking at 12 noon.
                    # 12 noon == 12 PM == hour 12
                    if dot_dict.hour != 12:
                        dot_dict.hour += 12
            elif group_key == 'M':
                dot_dict.minute = int(found_dict['M'])
            elif group_key == 'S':
                dot_dict.second = int(found_dict['S'])
            elif group_key == 'f':
                s = found_dict['f']
                # Pad to always return microseconds.
                s += "0" * (6 - len(s))
                dot_dict.fraction = int(s)
            elif group_key == 'A':
                dot_dict.weekday = locale_time.f_weekday.index(
                    found_dict['A'].lower())
            elif group_key == 'a':
                dot_dict.weekday = locale_time.a_weekday.index(
                    found_dict['a'].lower())
            elif group_key == 'w':
                dot_dict.weekday = int(found_dict['w'])
            elif group_key == 'u':
                dot_dict.weekday = int(found_dict['u'])
            elif group_key == 'j':
                dot_dict.julian = int(found_dict['j'])
            elif group_key in ('U', 'W'):
                dot_dict.week_of_year = int(found_dict[group_key])
            elif group_key == 'V':
                dot_dict.iso_week = int(found_dict['V'])
            elif group_key == 'z':
                z = found_dict['z']

                if z == 'Z':
                    dot_dict.gmtoff = 0
                else:
                    if z[3] == ':':
                        z = z[:3] + z[4:]

                        if len(z) > 5:
                            if z[5] != ':':
                                msg = ("Inconsistent use of : in "
                                       f"{found_dict['z']}")
                                raise ValueError(msg)

                            z = z[:5] + z[6:]

                    dot_dict.hours = int(z[1:3])
                    dot_dict.minutes = int(z[3:5])
                    dot_dict.seconds = int(z[5:7] or 0)
                    dot_dict.gmtoff = ((dot_dict.hours * 60 * 60) +
                                       (dot_dict.minutes * 60) +
                                       dot_dict.seconds)
                    gmtoff_remainder = z[8:]
                    # Pad to always return microseconds.
                    gmtoff_remainder_padding = (
                        "0" * (6 - len(gmtoff_remainder)))
                    dot_dict.gmtoff_fraction = int(gmtoff_remainder +
                                                   gmtoff_remainder_padding)

                    if z.startswith("-"):
                        dot_dict.gmtoff = -dot_dict.gmtoff
                        dot_dict.gmtoff_fraction = -dot_dict.gmtoff_fraction
            elif group_key == 'Z':
                # Since -1 is default value only need to worry about setting
                # tz if it can be something other than -1.
                found_zone = found_dict['Z'].lower()

                for value, tz_values in enumerate(locale_time.timezone):
                    if found_zone in tz_values:
                        # Deal with bad locale setup where timezone names are
                        # the same and yet time.daylight is true; too ambiguous
                        # to be able to tell what timezone has daylight savings
                        if (time.tzname[0] == time.tzname[1]and time.daylight
                            and found_zone not in (
                                "utc", "gmt")):  # pragma: no cover
                            break
                        else:
                            dot_dict.tz = value
                            break

        # Add timezone info
        dot_dict.tzname = found_dict.get("Z")
        return dot_dict

    def _check_iso_week(self, dot_dict):
        # Deal with the cases where ambiguities arise don't assume default
        # values for ISO week/year
        if dot_dict.iso_year is not None:
            if dot_dict.julian is not None:
                raise ValueError("Day of the year directive '%j' is not "
                                 "compatible with ISO year directive '%G'. "
                                 "Use '%Y' instead.")
            elif dot_dict.iso_week is None or dot_dict.weekday is None:
                raise ValueError("ISO year directive '%G' must be used with "
                                 "the ISO week directive '%V' and a weekday "
                                 "directive ('%A', '%a', '%w', or '%u').")
        elif dot_dict.iso_week is not None:
            if dot_dict.year is None or dot_dict.weekday is None:
                raise ValueError("ISO week directive '%V' must be used with "
                                 "the ISO year directive '%G' and a weekday "
                                 "directive ('%A', '%a', '%w', or '%u').")
            else:
                raise ValueError("ISO week directive '%V' is incompatible "
                                 "with the year directive '%Y'. Use the ISO "
                                 "year '%G' instead.")

    def _miscellaneous(self, dot_dict):
        if dot_dict.year is None:  # We need to make our best guess.
            if dot_dict.month == 0 and dot_dict.day == 5:
                dot_dict.year = 1  # 1 is first leap year of 1th century
            else:
                # Year 2 is not a leap year and is picked arbitrarily.
                dot_dict.year = 2

        # If we know the week of the year and what day of that week, we can
        # figure out the Julian day of the year.
        if dot_dict.julian is None and dot_dict.weekday is not None:
            if dot_dict.week_of_year is not None:
                dot_dict.julian = _calc_julian_from_U_or_W(
                    dot_dict.year, dot_dict.week_of_year, dot_dict.weekday)

                if dot_dict.julian is not None and dot_dict.julian <= 0:
                    bc = BahaiCalendar()
                    dot_dict.year -= 1
                    yday = 365 + bc._is_leap_year(dot_dict.year)
                    dot_dict.julian += yday
            elif (dot_dict.iso_year is not None
                  and dot_dict.iso_week is not None):
                datetime_result = datetime_date.fromisocalendar(
                    dot_dict.iso_year, dot_dict.iso_week, dot_dict.weekday + 1,
                    short=True)
                dot_dict.year = datetime_result.year
                dot_dict.month = datetime_result.month
                dot_dict.day = datetime_result.day

        if dot_dict.julian is None:
            # Cannot pre-calculate datetime_date() since can change in Julian
            # calculation and thus could have different value for the day of
            # the week calculation.
            # Need to add 1 to result since first day of the year is 1, not 0.
            dot_dict.julian = (datetime_date(dot_dict.year, dot_dict.month,
                                             dot_dict.day).toordinal() -
                               datetime_date(dot_dict.year, 1, 1).toordinal()
                               + 1)
        else:
            # Assume that if they bothered to include Julian day (or if it was
            # calculated above with year/week/weekday) it will be accurate.
            datetime_result = datetime_date.fromordinal(
                (dot_dict.julian - 1) +
                datetime_date(dot_dict.year, 1, 1).toordinal(), short=True)
            dot_dict.year = datetime_result.year
            dot_dict.month = datetime_result.month
            dot_dict.day = datetime_result.day

        if dot_dict.weekday is None:
            dot_dict.weekday = datetime_date(dot_dict.year, dot_dict.month,
                                             dot_dict.day).weekday()


def _strptime_time(data_string, format="%a %b %d %H:%M:%S %Y"):
    """
    Return a time struct based on the input string and the format string.
    """
    tt = StrpTime(data_string, format).start()[0]
    return _td_utils._build_struct_time(tt[:_td_utils._SHORT_STRUCT_TM_ITEMS],
                                        0, short_in=True)


def _strptime_datetime(cls, data_string, format="%a %b %d %H:%M:%S %Y"):
    """
    Return a datetime class cls instance based on the input string and the
    format string. (cls will always be a datetime class object)
    """
    tt, fraction, gmtoff_fraction = StrpTime(data_string, format).start()
    # *** TODO *** The below needs to be fixed when long date are implemented.
    tzname, gmtoff = tt[-2:]
    args = (*tt[:3], None, None, *tt[3: 6]) + (fraction,)

    if gmtoff is not None:
        tzdelta = datetime_timedelta(seconds=gmtoff,
                                     microseconds=gmtoff_fraction)

        if tzname:
            tz = datetime_timezone(tzdelta, tzname)
        else:
            tz = datetime_timezone(tzdelta)

        args += (tz,)

    return cls(*args)



# -*- coding: utf-8 -*-
#
# badidatetime/__init__.py
#
__docformat__ = "restructuredtext en"

import importlib
import geocoder
from tzlocal import get_localzone
from datetime import datetime as _dtime

from badidatetime.badi_calendar import BahaiCalendar
from badidatetime.gregorian_calendar import GregorianCalendar

dt_objects = ('date', 'datetime', 'time', 'timezone', 'timedelta', 'tzinfo',
              'MINYEAR', 'MAXYEAR', 'BADI_IANA', 'BADI_COORD', 'GMT_COORD',
              'UTC', 'BADI', 'LOCAL_COORD', 'LOCAL')


def _local_timezone_info():
    """
    Returns the offset in seconds, dst, IANA timezone key.

    :returns: Offset in seconds, True or False for dst, and the IANA key.
    :rtype: tuple

    .. note::

       Currently this must use the Python built in datetime, because the
       tzlocal package does not work completely with the badi datetime
       package.
    """
    localzone = get_localzone()
    dt = _dtime.now(localzone)
    offset = dt.utcoffset().total_seconds()
    dst = dt.dst().total_seconds() != 0
    return offset, dst, localzone.key


def _get_local_coordinates():
    """
    Get the locales coordinates and timezone offset.
    """
    offset, dst, key = _local_timezone_info()

    # Get latitude and longitude
    g = geocoder.ip('me')
    latitude = g.lat
    longitude = g.lng
    return latitude, longitude, offset / 3600


def enable_geocoder(enable=True):
    """
    Enable or disable the geocode query to find the local latitude, longitude,
    and zone. If this function is never run then the geocoder is nor run the
    the defaults for `datetime.LOCAL_COORD` and `datetime.LOCAL` as set to be
    local in Tehran Iran.

    :param bool enable: If True (default) geocoder is run else if False it is
                        not run.
    """
    badidt = importlib.import_module('badidatetime.datetime')

    if enable:
        badidt.LOCAL_COORD = _get_local_coordinates()
    else:
        badidt.LOCAL_COORD = badidt.BADI_COORD

    badidt.LOCAL = badidt.timezone.local = badidt.timezone._create(
        badidt.timedelta(hours=badidt.LOCAL_COORD[2]))
    badidt = importlib.reload(badidt)

    for obj in dt_objects:
        globals()[obj] = getattr(badidt, obj)


enable_geocoder(False)
__all__ = ('BahaiCalendar', 'GregorianCalendar', 'enable_geocoder')+dt_objects



# -*- coding: utf-8 -*-
#
# badidatetime/test/test_timedateutils.py
#
__docformat__ = "restructuredtext en"

import os
import sys
import locale
import importlib
import unittest
from unittest.mock import patch, PropertyMock

PWD = os.path.dirname(os.path.abspath(__file__))
BASE_DIR = os.path.dirname(os.path.dirname(PWD))
sys.path.append(BASE_DIR)

from .._timedateutils import TimeDateUtils

datetime = importlib.import_module('badidatetime.datetime')


class TestTimeDateUtils(unittest.TestCase):
    """
    Notes on running these tests. If you live outside of the USA you may need
    to install the USA local info on your computer or some tests will fail.
    There is no way to test for changing local information when most tests
    are hard coded.
    """

    def __init__(self, name):
        super().__init__(name)

    def setUp(self):
        self.locale_patcher = patch(
            'badidatetime._timedateutils.locale.nl_langinfo')
        self.mock_nl_langinfo = self.locale_patcher.start()

        def side_effect(item):
            if item == locale.LC_TIME:
                return 'en_US.UTF-8'
            elif item == locale.AM_STR:
                return 'AM'
            elif item == locale.PM_STR:
                return 'PM'
            elif item == locale.D_FMT:
                return '%m/%d/%Y'
            elif item == locale.T_FMT:
                return '%H:%M:%S'
            else:
                return 'Default Value'

        self.mock_nl_langinfo.side_effect = side_effect
        self._tdu = TimeDateUtils()

    def tearDown(self):
        self.locale_patcher.stop()

        #@unittest.skip("Temporarily skipped")
    def test__order_format(self):
        """
        Test that the _order_format returns a correctly parsed date or
        time format.
        """
        data = (
            ('%m/%d/%y', '%m/%d/%y', ['/', 'm', 'd', 'y']),
            ('%H:%M:%S', '%H:%M:%S', [':', 'H', 'M', 'S']),
            ('', '%m/%d/%y', ['/', 'm', 'd', 'y']),
            )
        msg = "Expected {}, with format {} and default {}, found {}."

        for fmt, default, expected_result in data:
            result = self._tdu._order_format(fmt, default)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, fmt, default, result))

    #@unittest.skip("Temporarily skipped")
    def test__find_time_order(self):
        """
        Test that the _find_time_order method parses the time format properly.
        """
        expected = '%I:%M:%S'
        result = self._tdu._find_time_order()
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_locale(self):
        """
        Test that the locale property is set correctly.
        """
        expected = "en_US.UTF-8"
        result = self._tdu.locale
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_am(self):
        """
        Test that the am property is set correctly
        """
        expected = 'AM'
        result = self._tdu.am
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_pm(self):
        """
        Test that the pm property is set correctly
        """
        expected = 'PM'
        result = self._tdu.pm
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_date_format(self):
        """
        Test that the date_format property is set correctly
        """
        expected = ['/', 'm', 'd', 'Y']
        result = self._tdu.date_format
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test_time_format(self):
        """
        Test that the time_format property is set correctly
        """
        expected = [':', 'I', 'M', 'S']
        result = self._tdu.time_format
        msg = f"Expected {expected}, found {result}."
        self.assertEqual(expected, result, msg)

    #@unittest.skip("Temporarily skipped")
    def test__checktm(self):
        """
        Test that the _checktm method does not raise an exception with
        an invalid tupple type.
        """
        bad_t, ttup_l, ttup_s, ttup_tl, ttup_ts = 0, 1, 2, 3, 4

        def make_ttup(t_type):
            if t_type == bad_t:
                ttup = date
            elif t_type == ttup_l:
                ttup = self._tdu._build_struct_time(date, dstflag)
            elif t_type == ttup_s:
                ttup = self._tdu._build_struct_time(date, dstflag,
                                                    short_in=True)
            else: # ttup_tl and ttup_ts
                ttup = date + (dstflag,)

            return ttup

        MIN_K = self._tdu.KULL_I_SHAY_MIN
        MAX_K = self._tdu.KULL_I_SHAY_MAX
        MIN_Y = self._tdu.MINYEAR
        MAX_Y = self._tdu.MAXYEAR
        err_msg0 = ("Invalid kull-i-shay {}, it must be in the range "
                    f"of [{MIN_K}, {MAX_K}].")
        err_msg1 = ("Invalid Váḥids '{}' in a Kull-i-Shay’, it must be in "
                    "the range of [1, 19].")
        err_msg2 = ("Invalid year '{}' in a Váḥid, it must be in the "
                    "range of [1, 19].")
        err_msg3 = ("Invalid year '{}' it must be in the range of ["
                    f"{MIN_Y}, {MAX_Y}].")
        err_msg4 = "Invalid month '{}', it must be in the range of [0, 19]."
        err_msg5 = ("Invalid day '{}' for month '{}', it must be in the "
                    "range of [1, {}].")
        err_msg6 = "Invalid hour '{}', it must be in the range of [0, 24]."
        err_msg7 = "Invalid minute '{}', it must be in the range of [0, 59]."
        err_msg8 = "Invalid second '{}', it must be in the range of [0, 61]."
        err_msg9 = "Invalid week day '{}', it must be in the range of [0, 6]."
        err_msg10 = ("Invalid day '{}' in year, it must be in the range of "
                     "[1, 366].")
        err_msg11 = "Invalid isdst '{}', it must be in the range of [-1, 1]."
        err_msg12 = "The ttup argument {} is not a proper tuple."
        err_msg13 = "Invalid timetuple, found length {}."
        data = (
            ### Valid tuples
            ((MIN_K, 18, 1, 1, 1, 1, 1, 1), -1, ttup_l, False, ''),
            ((1, 10, 10, 9, 6, 8, 45, 1), -1, ttup_l, False, ''),
            ((MAX_K, 5, 2, 19, 19, 1, 1, 1), -1, ttup_l, False, ''),
            ((MIN_Y, 1, 1, 0, 0, 0), -1, ttup_s, False, ''),
            ((181, 9, 6, 8, 45, 1), -1, ttup_s, False, ''),
            ((MAX_Y, 19, 19, 0, 0, 0), -1, ttup_s, False, ''),
            ((1, 10, 10, 1, 1, 0, 0, 0, 4, 1), -1, ttup_tl, False, ''),
            ((181, 1, 1, 0, 0, 0, 4, 1), -1, ttup_ts, False, ''),
            ### Invalid tuples
            # Long form NamedTuple errors
            ((-6, 1, 1, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg0.format(-6)),
            ((5, 1, 1, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg0.format(5)),
            ((1, 0, 1, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg1.format(0)),
            ((1, 20, 1, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg1.format(20)),
            ((1, 1, 0, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg2.format(0)),
            ((1, 1, 20, 1, 1, 1, 1, 1), -1, ttup_l, True, err_msg2.format(20)),
            # Long form standard tuple errors
            ((-6, 1, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg0.format(-6)),
            ((5, 1, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg0.format(5)),
            ((1, 0, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg1.format(0)),
            ((1, 20, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg1.format(20)),
            ((1, 1, 0, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg2.format(0)),
            ((1, 1, 20, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg2.format(20)),
            # Short for NamedTuple errors
            ((-1843, 1, 1, 0, 0, 0), -1, ttup_s, True, err_msg3.format(-1843)),
            ((1162, 1, 1, 0, 0, 0), -1, ttup_s, True, err_msg3.format(1162)),
            # Short for standard tuple errors
            ((-1843, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, True,
             err_msg3.format(-1843)),
            ((1162, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, True,
             err_msg3.format(1162)),
            # All tuple types use the same code for the month, day, hour,
            # minute, second, wday, yday, and isdst fields.
            # Month
            ((1, -1, 1, 0, 0, 0), -1, ttup_s, True, err_msg4.format(-1)),
            ((1, 20, 1, 0, 0, 0), -1, ttup_s, True, err_msg4.format(20)),
            # Day
            ((1, 1, 1, 1, -1, 0, 0, 0, 1, 1), -1, ttup_tl, True,
             err_msg5.format(-1, 1, 19)),
            ((1, 1, 1, 1, 20, 0, 0, 0, 1, 1), -1, ttup_tl, True,
             err_msg5.format(20, 1, 19)),
            # Day, leap year
            ((1, 0, 0, 0, 0, 0, 1, 1), -1, ttup_ts, True,
             err_msg5.format(0, 0, 5)),
            ((1, 0, 6, 0, 0, 0, 1, 1), -1, ttup_ts, True,
             err_msg5.format(6, 0, 5)),
            # Hour
            ((1, 1, 1, 1, 1, -1, 1, 1), -1, ttup_l, True, err_msg6.format(-1)),
            ((1, 1, 1, 1, 1, 25, 1, 1), -1, ttup_l, True, err_msg6.format(25)),
            # Minute
            ((1, 1, 1, 1, -1, 0), -1, ttup_s, True, err_msg7.format(-1)),
            ((1, 1, 1, 1, 60, 0), -1, ttup_s, True, err_msg7.format(60)),
            # Second
            ((1, 1, 1, 1, 1, 1, 1, -1, 1, 1), -1, ttup_tl, True,
             err_msg8.format(-1)),
            ((1, 1, 1, 1, 1, 1, 1, 62, 1, 1), -1, ttup_tl, True,
             err_msg8.format(62)),
            # Week day
            ((1, 1, 1, 0, 0, 0, -1, 1), -1, ttup_ts, True, err_msg9.format(-1)),
            ((1, 1, 1, 0, 0, 0, 7, 1), -1, ttup_ts, True, err_msg9.format(7)),
            # Day in year
            ((1, 1, 1, 1, 1, 1, 1, 1, 1, -1), -1, ttup_tl, True,
             err_msg10.format(-1)),
            ((1, 1, 1, 1, 1, 1, 1, 1, 1, 367), -1, ttup_tl, True,
             err_msg10.format(367)),
            # isdst (Daylight savings time)
            ((1, 1, 1, 1, 1, 1, 1, 1), -2, ttup_ts, True, err_msg11.format(-2)),
            ((1, 1, 1, 1, 1, 1, 1, 1), 2, ttup_ts, True, err_msg11.format(2)),
            # Proper tuple
            ([1, 1, 1, 1, 1, 1], -1, bad_t, True,
             err_msg12.format("<class 'list'>")),
            ((1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, True,
             err_msg13.format(13)),
            )

        for date, dstflag, t_type, validity, expected_result in data:
            if validity: # Invalid tests
                try:
                    ttup = make_ttup(t_type)
                    self._tdu._checktm(ttup)
                except (AssertionError, TypeError) as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    # Raise an error when an AssertionError is not raised.
                    raise AssertionError(
                        f"With date {date} an error was not raised.")
            else:  # Valid tests (Nothing to assert)
                ttup = make_ttup(t_type)
                self._tdu._checktm(ttup)

    #@unittest.skip("Temporarily skipped")
    @patch('badidatetime._timedateutils.TimeDateUtils.date_format',
           new_callable=PropertyMock)
    def test_strftime(self, mock_property):
        """
        Test that the strftime method returns the correct string.
        """
        mock_property.return_value = ['/', 'm', 'd', 'Y']
        ttup_l, ttup_s, ttup_tl, ttup_ts = 1, 2, 3, 4
        data = (
            ('%a', (1, 1, 1, 1, 1, 0, 0, 0), -1, ttup_l, None, 'Fiḍ'),
            ('%A', (1, 1, 1, 0, 0, 0), -1, ttup_s, None, 'Fiḍāl'),
            ('%b', (1, 1, 1, 1, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, 'Bah'),
            ('%B', (1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_ts, None, 'Bahá'),
            ('%c', (1, 1, 1, 1, 1, 3, 1, 1), -1, ttup_l, None,
             'Fiḍ Bah 01 03:01:01 1 01 01'),
            ('%c', (1, 1, 1, 3, 1, 1), -1, ttup_s, None,
             'Fiḍ Bah 01 03:01:01 0001'),
            ('%C', (1, 10, 10, 1, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, '01'),
            ('%C', (181, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '01'),
            ('%d', (1, 1, 1, 1, 8, 0, 0, 0), -1, ttup_l, None, '08'),
            ('%-d', (1, 1, 8, 0, 0, 0), -1, ttup_s, None, '8'),
            ('%D', (1, 10, 10, 19, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None,
             '19/01/81'),
            ('%e', (181, 1, 9, 0, 0, 0, 1, 1), -1, ttup_ts, None, ' 9'),
            ('%f', (1, 10, 10, 1, 1, 0, 0, 60.025), -1, ttup_l, None, '025000'),
            ('%G', (181, 1, 1, 0, 0, 0), -1, ttup_s, None, '0181'),
            ('%G', (self._tdu.MINYEAR, 1, 1, 0, 0, 0), -1, ttup_s, None,
             '-1842'),
            ('%h', (1, 10, 10, 2, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, 'Jal'),
            ('%H', (181, 1, 1, 3, 0, 0, 1, 1), -1, ttup_ts, None, '03'),
            ('%-H', (1, 10, 10, 1, 1, 3, 0, 0), -1, ttup_l, None, '3'),
            ('%I', (181, 1, 1, 13, 0, 0), -1, ttup_s, None, '01'),
            ('%j', (1, 10, 10, 2, 1, 0, 0, 0), -1, ttup_l, None, '020'),
            ('%-j', (1, 10, 10, 2, 1, 0, 0, 0), -1, ttup_l, None, '20'),
            ('%k', (181, 1, 1, 3, 0, 0, 1, 1), -1, ttup_ts, None, '03'),
            ('%:K', (1, 10, 10, 1, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, '1'),
            ('%:K', (181, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '1'),
            ('%l', (1, 10, 10, 1, 1, 13, 0, 0), -1, ttup_l, None, ' 1'),
            ('%-l', (181, 1, 1, 13, 0, 0), -1, ttup_s, None, '1'),
            ('%m', (1, 1, 1, 9, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, '09'),
            ('%-m', (1, 9, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '9'),
            ('%M', (1, 10, 10, 1, 1, 0, 9, 0), -1, ttup_l, None, '09'),
            ('%-M', (181, 1, 1, 0, 9, 0), -1, ttup_s, None, '9'),
            ('%n', (1, 1, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, None, '\n'),
            ('%p', (181, 1, 1, 9, 0, 0, 1, 1), -1, ttup_ts, None, 'AM'),
            ('%p', (181, 1, 1, 19, 0, 0, 1, 1), -1, ttup_ts, None, 'PM'),
            ('%r', (1, 1, 1, 1, 1, 13, 5, 2), -1, ttup_l, None, '01:05:02 PM'),
            ('%S', (1, 1, 1, 0, 0, 5), -1, ttup_s, None, '05'),
            ('%-S', (1, 1, 1, 1, 1, 0, 0, 5, 1, 1), -1, ttup_tl, None, '5'),
            ('%T', (181, 1, 1, 3, 10, 5, 1, 1), -1, ttup_ts, None, '03:10:05'),
            ('%u', (1, 10, 10, 1, 1, 0, 0, 0), -1, ttup_l, None, '4'),
            ('%U', (1, 1, 1, 0, 0, 0), -1, ttup_s, None, '01'),
            ('%U', (181, 1, 10, 0, 0, 0), -1, ttup_s, None, '02'),
            ('%V', (1, 10, 10, 2, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, '04'),
            ('%:V', (1, 10, 10, 1, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, '10'),
            ('%:V', (181, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '10'),
            ('%W', (1, 10, 10, 1, 1, 0, 0, 0), -1, ttup_l, None, '01'),
            ('%x', (181, 1, 1, 0, 0, 0), -1, ttup_s, None, '01/01/0181'),
            ('%X', (1, 1, 1, 1, 1, 1, 1, 1, 1, 1), -1, ttup_tl, None,
             '01:01:01'),
            ('%y', (1, 10, 10, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '81'),
            ('%y', (181, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '81'),
            ('%-y', (1, 1, 1, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '1'),
            ('%-y', (1, 1, 1, 0, 0, 0, 1, 1), -1, ttup_ts, None, '1'),
            ('%Y', (181, 11, 17, 0, 0, 0), -1, ttup_s, None, '0181'),
            ('%z', (1, 1, 1, 1, 1, 13, 5, 2, 1, 1), -1, ttup_tl, None, ''),
            ('%z', (1, 1, 1, 1, 1, 13, 5, 2), -1, ttup_l, datetime.BADI,
             '+0350'),
            ('%z', (1, 1, 1, 13, 5, 2), -1, ttup_s, datetime.BADI, '+0350'),
            ('%:z', (181, 1, 1, 13, 5, 2), -1, ttup_s, datetime.BADI, '+03:50'),
            ('%Z', (1, 10, 10, 2, 1, 0, 0, 0, 1, 1), -1, ttup_tl, None, ''),
            ('%Z', (1, 10, 10, 2, 1, 0, 0, 0), -1, ttup_l, datetime.BADI,
             'UTC+03:30'),
            ('%%', (181, 1, 1, 13, 5, 2, 1, 1), -1, ttup_ts, None, '%'),
            # Some composit formats
            ('%d/%m/%Y, %H:%M:%S', (1, 10, 10, 1, 1, 12, 30, 30), -1, ttup_l,
             None, '01/01/0181, 12:30:30'),
            ('%B %A %r', (181, 11, 16, 18, 40, 59), -1, ttup_s, None,
             'Mashíyyat Istijlāl 06:40:59 PM'),
            )
        msg = "Expected {}, with format {} and date {}. found {}."

        for fmt, date, dstflag, t_type, tz, expected_result in data:
            if t_type == ttup_l:
                ttup = self._tdu._build_struct_time(date, dstflag, tzinfo=tz)
            elif t_type == ttup_s:
                ttup = self._tdu._build_struct_time(date, dstflag, tzinfo=tz,
                                                    short_in=True)
            else: # ttup_tl and ttup_ts
                ttup = date + (dstflag,)

            result = self._tdu.strftime(fmt, ttup)
            self.assertEqual(expected_result, result, msg.format(
                    expected_result, fmt, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__check_format(self):
        """
        Test that the _check_format method does not raise an exception
        with an invalid format.
        """
        err_msg0 = "Invalid format character '{}'"
        err_msg1 = "Found an empty format string."
        data = (
            ('%c', False, None),
            ('%X', False, None),
            ('%P', True, err_msg0.format('%P')),
            ('%-P', True, err_msg0.format('%-P')),
            ('%:P', True, err_msg0.format('%:P')),
            ('', True, err_msg1),
            )
        msg = "Expected {}, with {}. found {}."

        for fmt, validity, expected_result in data:
            if validity:
                try:
                    result = self._tdu._check_format(fmt)
                except ValueError as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    result = result if result else None
                    raise AssertionError(f"With '{fmt}' an error is not "
                                         f"raised, with result {result}.")
            else:
                result = self._tdu._check_format(fmt)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, fmt, result))

    #@unittest.skip("Temporarily skipped")
    def test__find_midday(self):
        """
        Test that the _find_midday method returns the correct midday fraction
        of the day with either long or short form dates.
        """
        ttup_l, ttup_s = 1, 2
        data = (
            ((1, 10, 10, 11, 17, 9, 30, 30), -1, ttup_l, 0.49919999996200204),
            ((182, 1, 1, 18, 30, 30), -1, ttup_s, 0.5005864999257028),
            )
        msg = "Expected {}, with {}. found {}."

        for date, dstflag, t_type, expected_result in data:
            if t_type == ttup_l:
                ttup = self._tdu._build_struct_time(date, dstflag)
            else: # t_type == ttup_s
                ttup = self._tdu._build_struct_time(date, dstflag,
                                                    short_in=True)

            result = self._tdu._find_midday(ttup)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__get_year(self):
        """
        Test that the _get_year returns the year converted from a long
        form date or the year from a short date.
        """
        ttup_l, ttup_s = 1, 2
        data = (
            ((1, 10, 10, 11, 17, 9, 30, 30), -1, ttup_l, 181),
            ((182, 1, 1, 18, 30, 30), -1, ttup_s, 182),
            )
        msg = "Expected {}, with {}. found {}."

        for date, dstflag, t_type, expected_result in data:
            if t_type == ttup_l:
                ttup = self._tdu._build_struct_time(date, dstflag)
            else: # t_type == ttup_s
                ttup = self._tdu._build_struct_time(date, dstflag,
                                                    short_in=True)

            result = self._tdu._get_year(ttup)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__year_week_day(self):
        """
        Test that the _year_week_day return the year, week, and day of
        the week.
        """
        data = (
            # year, mon, day -> year, week, day of week
            ((181, 11, 17),    (181, 30, 7)),
            ((182, 1, 1),      (181, 53, 5)),
            ((183, 19, 19),    (183, 52, 7)),
            # Last week of a leap year where the beginning of the year
            # started on Fiḍāl.
            ((6, 19, 19),      (6, 53, 4)),
            ((12, 19, 19),     (12, 53, 4)),
            # Year where the first day of the year starts on Kamál and
            # is a leap year.
            ((174, 19, 17),    (175, 1, 1)),
            ((174, 19, 18),    (175, 1, 2)),
            ((174, 19, 19),    (175, 1, 3)),
            )
        msg = "Expected {} with date {}, found {}."

        for date, expected_result in data:
            result = self._tdu._year_week_day(*date)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__days_before_year(self):
        """
        Test that the _days_before_year function returns the correct
        number of days before the specified year.
        """
        data = (
            (-1842, 0),
            (-1841, 366), # Year -1842 was a leap year
            (-1840, 731),
            (181, 738886),
            )
        msg = "Expected {} with year {}, found {}."

        for year, expected_result in data:
            result = self._tdu._days_before_year(year)
            self.assertEqual(expected_result, result,
                             msg.format(expected_result, year, result))

    #@unittest.skip("Temporarily skipped")
    def test__days_in_month(self):
        """
        Test that the _days_in_month function returns the correct days
        in the specified month.
        """
        data = (
            (181, 1, 19),
            (181, 2, 19),
            (181, 3, 19),
            (181, 4, 19),
            (181, 5, 19),
            (181, 6, 19),
            (181, 7, 19),
            (181, 8, 19),
            (181, 9, 19),
            (181, 10, 19),
            (181, 11, 19),
            (181, 12, 19),
            (181, 13, 19),
            (181, 14, 19),
            (181, 15, 19),
            (181, 16, 19),
            (181, 17, 19),
            (181, 18, 19),
            (181, 0, 4),
            (181, 19, 19),
            (182, 0, 5),
            )
        msg = "Expected {} with year {} and month {}, found {}."

        for year, month, expected_result in data:
            result = self._tdu._days_in_month(year, month)
            self.assertEqual(expected_result, result,
                             msg.format(expected_result, year, month, result))

    #@unittest.skip("Temporarily skipped")
    def test__days_before_month(self):
        """
        Test that the _days_before_month function returns the correct
        number of days in the year proceedings the first day in the correct
        month.
        """
        data = (
            (181, 1, 0),
            (181, 2, 19),
            (181, 3, 38),
            (181, 4, 57),
            (181, 5, 76),
            (181, 6, 95),
            (181, 7, 114),
            (181, 8, 133),
            (181, 9, 152),
            (181, 10, 171),
            (181, 11, 190),
            (181, 12, 209),
            (181, 13, 228),
            (181, 14, 247),
            (181, 15, 266),
            (181, 16, 285),
            (181, 17, 304),
            (181, 18, 323),
            (181, 0, 342),
            (181, 19, 346),
            (183, 1, 0), # 182 is a leap year
            )
        msg = "Expected {} with year {} and month {}, found {}."

        for year, month, expected_result in data:
            result = self._tdu._days_before_month(year, month)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, year, month, result))

    #@unittest.skip("Temporarily skipped")
    def test__day_of_week(self):
        """
        Test that the _day_of_week function returns the correct day of
        the week (0 - 6) for a given year, month, and day.

        All days before 1752-09-14 (-91, 10, 8, 5, 46, 17.1264) in the
        Gregorian Calendar will seem wrong when compared to the Badi
        Calendar in UK and the US. This is when the Gregorian Calendar was
        adopted and compensated 11 days.
        """
        data = (
            ((-1842, 1, 1), 2), # 0001-03-19 (Saturday -> Jamál)
            ((-91, 9, 15), 6),  # 1752-09-02 (Wednesday -> `Idāl)
            ((-91, 10, 8), 4),  # 1752-09-14 (Thursday -> Istijlāl)
            ((1, 1, 1), 3),     # 1844-03-19 (Tuesday -> Fiḍāl)
            ((181, 9, 9), 2),   # 2024-08-26 (Monday -> Kamál)
            )
        msg = "Expected {} with date {}, found {}."

        for date, expected_result in data:
            result = self._tdu._day_of_week(*date)
            self.assertEqual(expected_result, result,
                             msg.format(expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__ymd2ord(self):
        """
        Test that the _ymd2ord function returns the correct number of days
        since Badi year -1842 including the current day.
        """
        data = (
            ((-1842, 1, 1), 78),
            ((-1841, 1, 1), 444),
            ((-1796, 1, 1), 16880),
            ((-1792, 1, 1), 18341),
            ((-1788, 1, 1), 19802),
            ((181, 1, 1), 738964),
            # 1st week of 181
            ((180, 19, 17), 738961),
            ((180, 19, 18), 738962),
            ((180, 19, 19), 738963),
            ((181, 1, 1), 738964),
            ((181, 1, 2), 738965),
            ((181, 1, 3), 738966),
            ((181, 1, 4), 738967),
            )
        msg = "Expected {} with date {}, found {}."

        for date, expected_result in data:
            result = self._tdu._ymd2ord(*date)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__ord2ymd(self):
        """
        Test that the _ord2ymd function returns the year, month, and day
        from the Badi year -1842.
        """
        data = (
            (78, False, (-5, 18, 1, 1, 1)),
            (78, True, (-1842, 1, 1)),
            (444, True, (-1841, 1, 1)),
            (673219, True, (0, 19, 19)),
            (673220, True, (1, 1, 1)),
            (738964, True, (181, 1, 1)),
            # 1st week of 181
            (738961, True, (180, 19, 17)),
            (738962, True, (180, 19, 18)),
            (738963, True, (180, 19, 19)),
            (738964, True, (181, 1, 1)),
            (738965, True, (181, 1, 2)),
            (738966, True, (181, 1, 3)),
            (738967, True, (181, 1, 4)),
            (739100, True, (181, 8, 4)),
            )
        msg = "Expected {} with ordinal {} and short {}, found {}."

        for ordinal, short, expected_result in data:
            result = self._tdu._ord2ymd(ordinal, short=short)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, ordinal, short, result))

    #@unittest.skip("Temporarily skipped")
    def test__build_struct_time(self):
        """
        Test that the _build_struct_time correctly builds a timetuple object.
        """
        data = (
            ((181, 9, 6, 8, 45, 1), -1, None, True,
             ("structures.ShortFormStruct(tm_year=181, tm_mon=9, tm_mday=6, "
              "tm_hour=8, tm_min=45, tm_sec=1, tm_wday=6, tm_yday=158, "
              "tm_isdst=-1)", None, None)),
            ((1, 1, 1, 0, 0, 0), -1, datetime.BADI, True,
             ("structures.ShortFormStruct(tm_year=1, tm_mon=1, tm_mday=1, "
              "tm_hour=0, tm_min=0, tm_sec=0, tm_wday=3, tm_yday=1, "
              "tm_isdst=-1)", 'UTC+03:30', 12600.0)),
            ((1, 10, 10, 9, 6, 8, 45, 1), -1, None, False,
             ("structures.LongFormStruct(tm_kull_i_shay=1, tm_vahid=10, "
              "tm_year=10, tm_mon=9, tm_mday=6, tm_hour=8, tm_min=45, "
              "tm_sec=1, tm_wday=6, tm_yday=158, tm_isdst=-1)", None, None)),
            ((1, 1, 1, 0, 0, 0), -1, datetime.BADI, True,
             ('structures.ShortFormStruct(tm_year=1, tm_mon=1, tm_mday=1, '
              'tm_hour=0, tm_min=0, tm_sec=0, tm_wday=3, tm_yday=1, '
              'tm_isdst=-1)', 'UTC+03:30', 12600.0)),
            )
        msg0 = "Expected {} with dt {}, dst {}, and timezone {}, found {}."
        msg1 = "Expected {}, found {}."

        for dt, dst, tz, short_in, expected_result in data:
            result = self._tdu._build_struct_time(dt, dst, tzinfo=tz,
                                                  short_in=short_in)
            self.assertEqual(expected_result[0], str(result), msg0.format(
                expected_result, dt, dst, tz, result))
            self.assertEqual(expected_result[1], result.tm_zone,
                             msg1.format(expected_result[1], result.tm_zone))
            self.assertEqual(expected_result[2], result.tm_gmtoff,
                             msg1.format(expected_result[2], result.tm_gmtoff))

    #@unittest.skip("Temporarily skipped")
    def test__isoweek_to_badi(self):
        """
        Test that the _isoweek_to_badi function returns the year, month,
        and day for the beginning of the week either before or after
        depending on what day in the week the day falls.
        """
        err_msg0 = "Invalid week: {}"
        err_msg1 = "Invalid weekday: {} (range is [1, 7])"
        data = (
            # year  week day
            ((-1842,  1,  1), False, False, (-5, 17, 19, 19, 18)),
            ((-1842,  1,  1), True,  False, (-1843, 19, 18)),
            ((    1, 52,  1), True, False, (1, 19, 15)),
            ((   46, 52,  1), True, False, (46, 19, 15)),
            ((  175, 52,  1), True,  False, (175, 19, 16)),
            ((  176, 52,  1), True,  False, (176, 19, 15)),
            ((  181,  1,  1), True,  False, (180, 19, 17)),
            ((  181,  1,  7), True,  False, (181, 1, 4)),
            ((  181, 20,  7), True,  False, (181, 8, 4)),
            ((  181, 52,  1), True,  False, (181, 19, 16)),
            ((  181, 52,  7), True,  False, (182, 1, 3)),
            ((  182, 52,  1), True,  False, (182, 19, 14)),
            ((  183,  1,  1), True,  False, (183, 1, 2)),
            ((  187, 52,  1), True, False, (187, 19, 15)),
            ((  192, 52,  1), True, False, (192, 19, 16)),
            ((  193, 52,  1), True, False, (193, 19, 15)),
            ((  194, 52,  1), True, False, (194, 19, 14)),
            ((  198, 52,  1), True, False, (198, 19, 16)),
            ((  199, 52,  1), True, False, (199, 19, 14)),
            ((  203, 52,  1), True, False, (203, 19, 16)),
            ((  204, 52,  1), True, False, (204, 19, 15)),
            ((  209, 52,  1), True, False, (209, 19, 16)),
            ((  210, 52,  1), True, False, (210, 19, 15)),
            ((  215, 52,  1), True, False, (215, 19, 16)),
            ((  216, 52,  1), True, False, (216, 19, 14)),
            ((  181, 53,  1), True, True, err_msg0.format(53)),
            ((  181, 20, 10), False, True, err_msg1.format(10)),
            )
        msg = "Expected {} with (year, week, day) {}, found {}."

        for item, short, validity, expected_result in data:
            if validity:
                with self.assertRaises(AssertionError) as cm:
                    self._tdu._isoweek_to_badi(*item)

                message = str(cm.exception)
                self.assertEqual(expected_result, message)
            else:
                result = self._tdu._isoweek_to_badi(*item, short=short)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, item, result))

    #@unittest.skip("Temporarily skipped")
    def test__isoweek1jalal(self):
        """
        Test that the _isoweek1jalal function returns the ordinal day number
        of the first week with more than 3 days in it.
        """
        data = (
            (  1, 673217), # 1844-03-16
            (181, 738961), # 2024-03-16
            (182, 739332), # 2025-03-22
            (183, 739696), # 2026-03-21
            )
        msg = "Expected {} with year {}, found {}."

        for year, expected_result in data:
            result = self._tdu._isoweek1jalal(year)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, year, result))

    #@unittest.skip("Temporarily skipped")
    def test__parse_isoformat_date_time_timezone(self):
        """
        Test that the _parse_isoformat_date_time_timezone function returns a
        parsed date and time ISO string.
        """
        data = (
            ('-18420101T120000', (-1842, 1, 1, 12, 0, 0), 'None'),
            ('-1842-01-01T12:00:00', (-1842, 1, 1, 12, 0, 0), 'None'),
            ('11610101T120000', (1161, 1, 1, 12, 0, 0), 'None'),
            ('1161-01-01T12:00:00', (1161, 1, 1, 12, 0, 0), 'None'),
            #('0181-W20T12:00:00', (181, 8, 17, 12, 0, 0), 'None'),
            ('0181-W20-5T12:00:00', (181, 8, 2, 12, 0, 0), 'None'),
            ('0001-01-01B', (1, 1, 1), 'UTC+03:30'),
            ('0001-01-01T00:00:00.0+03:30', (1, 1, 1, 0, 0, 0), 'UTC+03:30'),
            ('-0126-16-02T07:58:31.4976Z', (-126, 16, 2, 7, 58, 31.4976),
             'UTC'),
            ('0181-13-09+02', (181, 13, 9), 'UTC+02:00'),
            ('0181-13-09-05', (181, 13, 9), 'UTC-05:00'),
            )
        msg = "Expected {} with dtstr {}, found {}."

        for dtstr, expected_result0, expected_result1 in data:
            date, time, tz = self._tdu._parse_isoformat_date_time_timezone(
                dtstr)
            self.assertEqual(expected_result0, date + time, msg.format(
                expected_result0, dtstr, date + time))
            self.assertEqual(expected_result1, str(tz), msg.format(
                expected_result1, dtstr, tz))

    #@unittest.skip("Temporarily skipped")
    def test__parse_isoformat_date(self):
        """
        Test that the _parse_isoformat_date function parses the date
        correctly from ISO standard formats.
        """
        err_msg0 = "Invalid character {} in incoming date string."
        err_msg1 = "Year is out of range: {}, min {}, max {}."
        err_msg2 = ("Invalid format, there must be between 0 to 2 hyphens "
                    "(-) in the date format or there can be one uppercase "
                    "(W) week identifier and between 0 and 2 hyphens (-) "
                    "used.")
        err_msg3 = "Invalid ISO string {}."
        data = (
            ('', False, ()),
            ('0181-01', False, (181, 1, 1)),
            ('01810101', False, (181, 1, 1)),
            ('0181-01-01', False, (181, 1, 1)),
            ('0181W01', False, (180, 19, 17)),
            ('0181-W01', False, (180, 19, 17)),
            ('0181-W01-1', False, (180, 19, 17)),
            ('0181-W01-2', False, (180, 19, 18)),
            ('0181-W01-3', False, (180, 19, 19)),
            ('0181-W01-4', False, (181, 1, 1)),
            ('0181-W01-5', False, (181, 1, 2)),
            ('0181-W01-6', False, (181, 1, 3)),
            ('0181-W01-7', False, (181, 1, 4)),
            ('0181W017', False, (181, 1, 4)),
            ('0181W20', False, (181, 7, 17)),
            ('0181W207', False, (181, 8, 4)),
            ('0181001', False, (181, 1, 1)),
            ('0181019', False, (181, 1, 19)),
            ('0181324', False, (181, 18, 1)),
            ('0181342', False, (181, 18, 19)),
            ('0181343', False, (181, 0, 1)),
            ('0181346', False, (181, 0, 4)),
            ('0181347', False, (181, 19, 1)),
            ('0181365', False, (181, 19, 19)),
            ('0181-001', False, (181, 1, 1)),
            ('0181-019', False, (181, 1, 19)),
            ('0182-324', False, (182, 18, 1)),
            ('0182-342', False, (182, 18, 19)),
            ('0182-343', False, (182, 0, 1)),
            ('0182-347', False, (182, 0, 5)),
            ('0182-348', False, (182, 19, 1)),
            ('0182-366', False, (182, 19, 19)),
            ('0001/01/01', True, err_msg0.format("'/'")),
            ('015s', True, err_msg0.format("'s'")),
            ('-1843-01', True, err_msg1.format(-1843, self._tdu.MINYEAR,
                                               self._tdu.MAXYEAR)),
            ('1162-01', True, err_msg1.format(1162, self._tdu.MINYEAR,
                                              self._tdu.MAXYEAR)),
            ('0181-01-01-', True, err_msg2),
            ('0181-W10-1-', True, err_msg2),
            ('0181-W101', True, err_msg3.format('0181-W101')),
            )
        msg = "Expected {} with ISO date {}, found {}."

        for date, validity, expected_result in data:
            if validity:
                try:
                    result = self._tdu._parse_isoformat_date(date)
                except (AssertionError, ValueError, IndexError) as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    result = result if result else None
                    raise AssertionError(f"With '{date}' an error is not "
                                         f"raised, with result {result}.")
            else:
                result = self._tdu._parse_isoformat_date(date)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, date, result))

    #@unittest.skip("Temporarily skipped")
    def test__parse_isoformat_time(self):
        """
        Test that the _parse_isoformat_time function parses the time
        correctly from ISO standard formats.
        """
        err_msg0 = "Invalid character {} in incoming time string."
        err_msg1 = ("Cannot have both a 'T' and a space or more than one "
                    "of either to indicate time.")
        err_msg2 = ("Invalid time string, 1st character must be one of ( T), "
                    "found {}")
        err_msg3 = "Invalid number of colons (:), can be 0 - 2, found {}"
        err_msg4 = "Invalid number of dots (.), can be 0 - 1, found {}"
        err_msg5 = "Invalid time string, found {}"
        data = (
            ('', False, ()),
            ('T14', False, (14, 0, 0)),
            (' 14', False, (14, 0, 0)),
            ('T14.2', False, (14, 12, 0)),
            ('T1412', False, (14, 12, 0)),
            ('T1412.45', False, (14, 12, 27)),
            ('T141232', False, (14, 12, 32)),
            ('T141232.029', False, (14, 12, 32.029)),
            ('T14:12', False, (14, 12, 0)),
            ('T14:12.45', False, (14, 12, 27)),
            ('T14:12:32', False, (14, 12, 32)),
            ('T14:12:32.029', False, (14, 12, 32.029)),
            ('S01:01:01', True, err_msg0.format("'S'")),
            ('T ', True, err_msg1),
            ('TT', True, err_msg1),
            ('  ', True, err_msg1),
            ('01:01:01', True, err_msg2.format("'01:01:01'")),
            ('T:::', True, err_msg3.format(3)),
            ('T..', True, err_msg4.format(2)),
            ('T014.2', True, err_msg5.format("'T014.2'")),
            )
        msg = "Expected {} with ISO time {}, found {}."

        for time, validity, expected_result in data:
            if validity:
                try:
                    result = self._tdu._parse_isoformat_time(time)
                except (AssertionError, ValueError) as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    result = result if result else None
                    raise AssertionError(f"With {time} an error is not "
                                         f"raised, with result {result}.")
            else:
                result = self._tdu._parse_isoformat_time(time)
                self.assertEqual(expected_result, result, msg.format(
                    expected_result, time, result))

    #@unittest.skip("Temporarily skipped")
    def test__parse_isoformat_timezone(self):
        """
        Test that the _parse_isoformat_timezone
        """
        err_msg0 = "Invalid character {} in incoming timezone string."
        err_msg1 = ("Can only have one of (-+Z) and no more than one of (-+Z) "
                    "to indicate a timezone.")
        err_msg2 = ("Invalid timezone string, 1st character must be one of "
                    "(-+Z), found {}")
        err_msg3 = "Invalid number of colons (:), can be 0 - 1, found {}"
        data = (
            ('', False, 'None'),
            ('Z', False, 'UTC'),
            ('B', False, 'UTC+03:30'),
            ('+05', False, 'UTC+05:00'),
            ('-05', False, 'UTC-05:00'),
            ('A', True, err_msg0.format("'A'")),
            ('-+0', True, err_msg1),
            ('--0', True, err_msg1),
            ('03:30', True, err_msg2.format("'03:30'")),
            ('-05:30:30', True, err_msg3.format(2)),
            )
        msg = "Expected {} with ISO timezone {}, found {}."

        for timezone, validity, expected_result in data:
            if validity:
                try:
                    result = self._tdu._parse_isoformat_timezone(timezone)
                except (AssertionError, ValueError) as e:
                    self.assertEqual(expected_result, str(e))
                else:
                    result = result if result else None
                    raise AssertionError(f"With {timezone} an error is not "
                                         f"raised, with result {result}.")
            else:
                result = self._tdu._parse_isoformat_timezone(timezone)
                self.assertEqual(expected_result, str(result), msg.format(
                    expected_result, timezone, result))

    #@unittest.skip("Temporarily skipped")
    def test__check_date_fields(self):
        """
        Test that the _check_date_fields function correctly raises
        assertion exceptions.

        A more complete test is in badidatetime/tests/test_badi_calendar.py.
        """
        err_msg0 = ("Invalid Váḥids '{}' in a Kull-i-Shay’, it must be in "
                    "the range of [1, 19].")
        err_msg1 = ("Invalid year '{}' in a Váḥid, it must be in the range "
                    "of [1, 19].")
        err_msg2 = "Invalid month '{}', it must be in the range of [0, 19]."
        err_msg3 = ("Invalid day '{}' for month '{}', it must be in the "
                    "range of [1, {}].")
        data = (
            # Valid short form Badi dates
            ((self._tdu.MINYEAR, 1, 1), False, ''),
            ((self._tdu.MAXYEAR, 1, 1), False, ''),
            # Invalid Váḥid
            ((1, 0, 1, 1, 1), True, err_msg0.format(0)),
            ((1, 20, 1, 1, 1), True, err_msg0.format(20)),
            # Invalid year
            ((1, 10, 0, 1, 1), True, err_msg1.format(0)),
            ((1, 10, 20, 1, 1), True, err_msg1.format(20)),
            # Invalid month
            ((1, 10, 10, -1, 1), True, err_msg2.format(-1)),
            ((1, 10, 10, 20, 1), True, err_msg2.format(20)),
            # Invalid Ayyám-i-Há day
            ((1, 10, 3, 0, 0), True, err_msg3.format(0, 0, 5)),
            ((1, 10, 3, 0, 6), True, err_msg3.format(6, 0, 5)),
            # Invalid normal day
            ((1, 10, 3, 2, 0), True, err_msg3.format(0, 2, 19)),
            ((1, 10, 3, 2, 20), True, err_msg3.format(20, 2, 19)),
            # Test short form date.
            ((181, 20, 1), True, err_msg2.format(20)),
            )

        for date, validity, err_msg in data:
            short_in = False if len(date) == 5 else True

            if validity:
                try:
                    self._tdu._check_date_fields(*date, short_in=short_in)
                except AssertionError as e:
                    self.assertEqual(err_msg, str(e))
                else:
                    raise AssertionError(f"date {date}")
            else:
                self._tdu._check_date_fields(*date, short_in=short_in)

    #@unittest.skip("Temporarily skipped")
    def test__check_time_fields(self):
        """
        Test that the _check_time_fields function correctly raises
        assertion exceptions.
        """
        err_msg0 = "Invalid hour '{}', it must be in the range of [0, 24]."
        err_msg1 = "Invalid minute '{}', it must be in the range of [0, 59]."
        err_msg2 = "Invalid second '{}', it must be in the range of [0, 61]."
        err_msg3 = ("Invalid microseconds '{}', it must be in the range of "
                    "[0, 999999].")
        err_msg4 = "The fold argument '{}' must be either 0 or 1."
        data = (
            ((0, 0, 0, 0, 0), False, ''),
            ((24, 59, 59, 999999, 1), False, ''),
            ((-1, 0, 0, 0, 0), True, err_msg0.format(-1)),
            ((25, 0, 0, 0, 0), True, err_msg0.format(25)),
            ((0, -1, 0, 0, 0), True, err_msg1.format(-1)),
            ((0, 60, 0, 0, 0), True, err_msg1.format(60)),
            ((0, 0, -1, 0, 0), True, err_msg2.format(-1)),
            ((0, 0, 61, 0, 0), True, err_msg2.format(61)),
            ((0, 0, 0, -1, 0), True, err_msg3.format(-1)),
            ((0, 0, 0, 1000000, 0), True, err_msg3.format(1000000)),
            ((0, 0, 0, 0, -1), True, err_msg4.format(-1)),
            ((0, 0, 0, 0, 2), True, err_msg4.format(2)),
            )
        for time, validity, err_msg in data:
            if validity:
                try:
                    self._tdu._check_time_fields(*time)
                except AssertionError as e:
                    self.assertEqual(err_msg, str(e))
                else:
                    raise AssertionError(f"time {time}")
            else:
                self._tdu._check_time_fields(*time)

    #@unittest.skip("Temporarily skipped")
    def test__wrap_strftime(self):
        """
        Test that the _wrap_strftime function returns a formatted time
        string.
        """
        d_t, dt_t = 1, 2
        data = (
            ((181, 1, 1), '%d/%m/%Y, %H:%M:%S', d_t, '01/01/0181, 00:00:00'),
            ((1, 1, 8), '%-d', d_t, '8'),
            #((181, 1, 1, 0, 9, 0), dt_t, '%-M', ''),
            )
        msg = "Expected {} with date {} and format {}, found {}."

        for date, fmt, obj_t, expected_result in data:
            if obj_t == d_t:
                d = datetime.datetime(*date)
            else:
                d = datetime(*date)

            tt = d.timetuple()
            result = self._tdu._wrap_strftime(date, fmt, tt)
            self.assertEqual(expected_result, result, msg.format(
                expected_result, date, format, result))
